from typing import Any

import pytest

from across.tools import Coordinate, Polygon
from across.tools.core.schemas import BaseSchema
from across.tools.footprint import Footprint


@pytest.fixture
def simple_polygon() -> Polygon:
    """
    Returns a very simple polygon center at 0, 0 with side = 1.0
    """
    coordinates = [
        Coordinate(ra=-0.5, dec=0.5),
        Coordinate(ra=0.5, dec=0.5),
        Coordinate(ra=0.5, dec=-0.5),
        Coordinate(ra=-0.5, dec=-0.5),
        Coordinate(ra=-0.5, dec=0.5),
    ]
    return Polygon(coordinates=coordinates)


@pytest.fixture
def simple_footprint(simple_polygon: Polygon) -> Footprint:
    """
    Instantiates  a simple footprint from a simple polygon
    """
    return Footprint(detectors=[simple_polygon])


@pytest.fixture(params=["bad detector", 42, [42, 42]])
def invalid_detector(request: pytest.FixtureRequest) -> Any:
    """
    Parameters to be passed into the projection tests
    """
    return request.param


@pytest.fixture
def origin_coordinate() -> Coordinate:
    """
    Instantiates a coordinate at ra=0, dec=0
    """
    return Coordinate(ra=0, dec=0)


@pytest.fixture(params=["bad roll_angle", -361, 361])
def invalid_roll_angle(request: pytest.FixtureRequest) -> Any:
    """
    Parameters to be passed into the projection tests
    """
    return request.param


@pytest.fixture
def ra45_dec45_coordinate() -> Coordinate:
    """
    Instantiates a coordinate at ra=45, dec=45
    """
    return Coordinate(ra=45, dec=45)


def simple_footprint_projection_ra45_dec0_pos0() -> Footprint:
    """
    Instantiates a precalculated projected simple footprint at ra=45, dec=0, roll=0
    """
    return Footprint(
        detectors=[
            Polygon(
                coordinates=[
                    Coordinate(ra=44.5, dec=0.5),
                    Coordinate(ra=45.5, dec=0.5),
                    Coordinate(ra=45.5, dec=-0.5),
                    Coordinate(ra=44.5, dec=-0.5),
                    Coordinate(ra=44.5, dec=0.5),
                ]
            ),
        ]
    )


def simple_footprint_projection_ra0_dec45_pos0() -> Footprint:
    """
    Instantiates a precalculated projected simple footprint at ra=0, dec=45, roll=0
    """
    return Footprint(
        detectors=[
            Polygon(
                coordinates=[
                    Coordinate(ra=359.28669, dec=45.4978),
                    Coordinate(ra=0.71331, dec=45.4978),
                    Coordinate(ra=0.70097, dec=44.49784),
                    Coordinate(ra=359.29903, dec=44.49784),
                    Coordinate(ra=359.28669, dec=45.4978),
                ]
            )
        ]
    )


def simple_footprint_projection_ra0_dec0_pos45() -> Footprint:
    """
    Instantiates a precalculated projected simple footprint at ra=0, dec=0, roll=45
    """
    return Footprint(
        detectors=[
            Polygon(
                coordinates=[
                    Coordinate(ra=359.2929, dec=1e-05),
                    Coordinate(ra=359.99999, dec=0.7071),
                    Coordinate(ra=0.7071, dec=0.0),
                    Coordinate(ra=0.0, dec=-0.7071),
                    Coordinate(ra=359.2929, dec=0.0),
                ]
            ),
        ]
    )


class PrecalculatedProjections(BaseSchema):
    """
    Class to represent a pre-calculated projection
    """

    coordinate: Coordinate
    roll_angle: float
    projection: Footprint


@pytest.fixture(
    params=[
        PrecalculatedProjections(
            coordinate=Coordinate(ra=45, dec=0),
            roll_angle=0,
            projection=simple_footprint_projection_ra45_dec0_pos0(),
        ),
        PrecalculatedProjections(
            coordinate=Coordinate(ra=0, dec=45),
            roll_angle=0,
            projection=simple_footprint_projection_ra0_dec45_pos0(),
        ),
        PrecalculatedProjections(
            coordinate=Coordinate(ra=0, dec=0),
            roll_angle=45,
            projection=simple_footprint_projection_ra0_dec0_pos45(),
        ),
    ]
)
def precalculated_projections(request: pytest.FixtureRequest) -> Any:
    """
    Parameters to be passed into the projection tests
    """
    return request.param


@pytest.fixture()
def precalculated_hp_query_polygon() -> list[int]:
    """
    Precalculated hp.query_polygon for a simple footprint projected to Coordinate(45, 45)
    """
    return [
        197185,
        197187,
        197188,
        197189,
        197190,
        197191,
        197196,
        197197,
        197199,
        197200,
        197201,
        197202,
        197203,
        197204,
        197205,
        197206,
        197207,
        197208,
        197209,
        197210,
        197211,
        197212,
        197213,
        197214,
        197215,
        197232,
        197233,
        197235,
        197236,
        197237,
        197238,
        197239,
        197244,
        197245,
        196831,
        196847,
        196851,
        196852,
        196853,
        196854,
        196855,
        196856,
        196857,
        196858,
        196859,
        196860,
        196861,
        196862,
        196863,
        197376,
        197377,
        197378,
        197379,
        197380,
        197381,
        197382,
        197383,
        197384,
        197385,
        197386,
        197387,
        197388,
        197389,
        197390,
        197392,
        197393,
        197394,
        197408,
        197409,
        197410,
        196994,
        196995,
        197000,
        197001,
        197002,
        197003,
        197004,
        197006,
        197007,
        197024,
        197025,
        197026,
        197027,
        197028,
        197029,
        197030,
        197031,
        197032,
        197033,
        197034,
        197035,
        197036,
        197037,
        197038,
        197039,
        197040,
        197042,
        197043,
        197048,
        197049,
        197050,
        197051,
        197052,
        197054,
    ]


@pytest.fixture(params=["bad healpix_order", -5, 15])
def invalid_healpix_order(request: pytest.FixtureRequest) -> Any:
    """
    Parameters to be passed into the projection tests
    """
    return request.param



from typing import Literal

import numpy as np
import numpy.typing as npt
from astropy.coordinates import GeocentricTrueEcliptic, SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]
from pydantic import Field

from ...core.enums import ConstraintType
from ...ephemeris import Ephemeris
from .base import ConstraintABC


class EclipticLatitudeConstraint(ConstraintABC):
    """
    Constraint that avoids observing at low ecliptic latitudes.

    The ecliptic plane contains zodiacal dust that scatters sunlight, creating
    zodiacal light that can contaminate observations. This constraint ensures
    observations are conducted at sufficient ecliptic latitude to avoid this
    background light.

    Parameters
    ----------
    min_latitude : float
        Minimum ecliptic latitude (degrees) required for observation.
        Observations closer to the ecliptic plane than this will be constrained.

    Methods
    -------
    __call__(time, ephemeris, coordinate)
        Checks if the coordinate is too close to the ecliptic plane.
    """

    name: Literal[ConstraintType.ECLIPTIC_LATITUDE] = ConstraintType.ECLIPTIC_LATITUDE
    short_name: Literal["Ecliptic Latitude"] = "Ecliptic Latitude"
    min_latitude: float = Field(
        default=15.0, ge=0, le=90, description="Minimum ecliptic latitude (degrees) for valid observations"
    )

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> npt.NDArray[np.bool_]:
        """
        Check if the coordinate is too close to the ecliptic plane.

        Parameters
        ----------
        time : Time
            The time (not used for ecliptic coordinates).
        ephemeris : Ephemeris
            The ephemeris (not used for ecliptic coordinates).
        coordinate : SkyCoord
            The coordinate to check.

        Returns
        -------
        npt.NDArray[np.bool_]
            Boolean array where True indicates the coordinate violates the constraint
            (is too close to the ecliptic plane).
        """
        # Convert to ecliptic coordinates
        ecliptic_coord = coordinate.transform_to(GeocentricTrueEcliptic())

        # Check if the absolute ecliptic latitude is less than the minimum
        # This creates a "zone of avoidance" around the ecliptic plane
        in_constraint: npt.NDArray[np.bool_] = np.abs(ecliptic_coord.lat.deg) < self.min_latitude

        return in_constraint



from .constraint_type import ConstraintType
from .depth_unit import DepthUnit
from .energy_unit import EnergyUnit
from .frequency_unit import FrequencyUnit
from .solar_system_object import SolarSystemObject
from .twilight_type import TwilightType
from .wavelength_unit import WavelengthUnit

__all__ = [
    "DepthUnit",
    "EnergyUnit",
    "FrequencyUnit",
    "WavelengthUnit",
    "ConstraintType",
    "TwilightType",
    "SolarSystemObject",
]



from .base import Visibility
from .constraints_constructor import Constraint, Constraints, constraints_from_json, constraints_to_json
from .ephemeris_visibility import EphemerisVisibility, compute_ephemeris_visibility
from .joint_visibility import JointVisibility, compute_joint_visibility

__all__ = [
    "EphemerisVisibility",
    "JointVisibility",
    "Visibility",
    "constraints_from_json",
    "constraints_to_json",
    "compute_ephemeris_visibility",
    "compute_joint_visibility",
    "Constraint",
    "Constraints",
]



from .bandpass import EnergyBandpass, FrequencyBandpass, WavelengthBandpass, convert_to_wave
from .base import BaseSchema
from .coordinate import Coordinate
from .custom_types import AstropyDateTime, AstropyTimeDelta
from .healpix_order import HealpixOrder
from .polygon import Polygon
from .roll_angle import RollAngle
from .visibility import ConstrainedDate, ConstraintReason, VisibilityWindow, Window

__all__ = [
    "Coordinate",
    "Polygon",
    "BaseSchema",
    "RollAngle",
    "HealpixOrder",
    "EnergyBandpass",
    "WavelengthBandpass",
    "FrequencyBandpass",
    "convert_to_wave",
    "VisibilityWindow",
    "ConstrainedDate",
    "Window",
    "ConstraintReason",
    "AstropyDateTime",
    "AstropyTimeDelta",
]



import numpy as np
import pytest

from across.tools import EnergyBandpass, FrequencyBandpass, WavelengthBandpass, convert_to_wave, enums


class TestBandpassSchema:
    """
    Class that tests the functionality of the bandpass schemas
    """

    class TestWavelengthBandpass:
        """
        Class that tests the instantiation of the WavelengthBandpass schema
        """

        @pytest.mark.parametrize(
            "min, max",
            [
                (None, 1),
                (1, None),
                (-1, 1),
                (1, -1),
                (2, 1),
            ],
        )
        def test_should_throw_value_error_min_max(self, min: float | None, max: float | None) -> None:
            """
            Parameterized tests that validate bad instantiations for WavelengthBandpass.
                Should raise ValueError when min=None, max=None
                Should raise ValueError when min < 0
                Should raise ValueError when max < 0
                Should raise ValueError when min > max
            """
            with pytest.raises(ValueError):
                WavelengthBandpass(min=min, max=max, unit=enums.WavelengthUnit.ANGSTROM)

        @pytest.mark.parametrize(
            "central_wavelength, bandwidth",
            [(None, None), (None, 1), (1, None), (-1, 1), (1, -1), (1, 2), (1, 1)],
        )
        def test_should_throw_value_error_central_wavelength_bandwidth(
            self, central_wavelength: float | None, bandwidth: float | None
        ) -> None:
            """
            Parameterized tests that validate bad instantiations for WavelengthBandpass.
                Should raise ValueError when central_wavelength=None, bandwidth=None
                Should raise ValueError when central_wavelength < 0
                Should raise ValueError when bandwidth < 0
                Should raise ValueError when bandwidth >= central_wavelength
            """
            with pytest.raises(ValueError):
                WavelengthBandpass(
                    central_wavelength=central_wavelength,
                    bandwidth=bandwidth,
                    unit=enums.WavelengthUnit.ANGSTROM,
                )

        @pytest.mark.parametrize(
            "min, max, unit, central_wavelength, bandwidth",
            [
                (5000.0, 6000.0, enums.WavelengthUnit.ANGSTROM, 5500.0, 500.0),
                (500, 600, enums.WavelengthUnit.NANOMETER, 5500, 500),
                (0.5, 0.6, enums.WavelengthUnit.MICRON, 5500, 500),
                (0.0005, 0.0006, enums.WavelengthUnit.MILLIMETER, 5500, 500),
            ],
        )
        def test_should_convert_min_max_to_correct_central_wavelength_bandwidth(
            self,
            min: float,
            max: float,
            unit: enums.WavelengthUnit,
            central_wavelength: float,
            bandwidth: float,
        ) -> None:
            """
            Parameterized test that confirms whether or not WavelengthBandpass converts
            the appropriate min max to central wavelength and bandpass.
            """
            wavelength_bandpass = WavelengthBandpass(min=min, max=max, unit=unit)
            assert all(
                [
                    np.isclose(wavelength_bandpass.central_wavelength, central_wavelength, 0.001),  # type:ignore
                    np.isclose(wavelength_bandpass.bandwidth, bandwidth, 0.001),  # type:ignore
                ]
            )

    class TestEnergyBandpass:
        """
        Class that tests the instantiation of the EnergyBandpass schema
        """

        @pytest.mark.parametrize(
            "min, max",
            [
                (None, None),
                (None, 1),
                (1, None),
                (-1, 1),
                (1, -1),
            ],
        )
        def test_should_throw_value_error_min_max(self, min: float | None, max: float | None) -> None:
            """
            Parameterized tests that validate bad instantiations for EnergyBandpass.
                Should raise ValueError when min=None, max=None
                Should raise ValueError when min < 0
                Should raise ValueError when max < 0
            """
            with pytest.raises(ValueError):
                EnergyBandpass(min=min, max=max, unit=enums.EnergyUnit.eV)

    class TestFrequencyBandpass:
        """
        Class that tests the instantiation of the FrequencyBandpass schema
        """

        @pytest.mark.parametrize(
            "min, max",
            [
                (None, None),
                (None, 1),
                (1, None),
                (-1, 1),
                (1, -1),
            ],
        )
        def test_should_throw_value_error_min_max(self, min: float | None, max: float | None) -> None:
            """
            Parameterized tests that validate bad instantiations for FrequencyBandpass.
                Should raise ValueError when min=None, max=None
                Should raise ValueError when min < 0
                Should raise ValueError when max < 0
            """
            with pytest.raises(ValueError):
                FrequencyBandpass(min=min, max=max, unit=enums.FrequencyUnit.Hz)

    class TestConvertToWave:
        """
        Class that tests the bandpass.convert_to_wave method
        """

        @pytest.mark.parametrize(
            "min, max, unit, min_wave, max_wave",
            [
                (1, 10, enums.EnergyUnit.eV, 1239.842, 12398.420),
                (1, 10, enums.EnergyUnit.keV, 1.23984198, 12.3984198),
                (1, 10, enums.EnergyUnit.MeV, 1.23984198e-3, 1.23984198e-2),
                (1, 10, enums.EnergyUnit.GeV, 1.23984198e-6, 1.23984198e-5),
                (1, 10, enums.EnergyUnit.TeV, 1.23984198e-9, 1.23984198e-8),
            ],
        )
        def test_convert_energy_to_wavelength(
            self, min: float, max: float, unit: enums.EnergyUnit, min_wave: float, max_wave: float
        ) -> None:
            """
            Parameterized test that confirms that the convert_to_wave method
            converts Energy min/max to appropriate Wavelength min/max
            """
            energy_bandpass = EnergyBandpass(min=min, max=max, unit=unit)
            wavelength_bandpass = convert_to_wave(energy_bandpass)
            assert all(
                [
                    np.isclose(wavelength_bandpass.min, min_wave),  # type:ignore
                    np.isclose(wavelength_bandpass.max, max_wave),  # type:ignore
                ]
            )

        @pytest.mark.parametrize(
            "min, max, unit, min_wave, max_wave",
            [
                (1, 10, enums.FrequencyUnit.Hz, 2.997924e17, 2.997924e18),
                (1, 10, enums.FrequencyUnit.kHz, 299792457999999.9, 2997924579999999.5),
                (1, 10, enums.FrequencyUnit.MHz, 299792457999.9, 2997924579999.9),
                (1, 10, enums.FrequencyUnit.GHz, 299792457.9, 2997924580),
                (1, 10, enums.FrequencyUnit.THz, 299792.4579, 2997924.579),
            ],
        )
        def test_convert_frequency_to_wavelength(
            self, min: float, max: float, unit: enums.FrequencyUnit, min_wave: float, max_wave: float
        ) -> None:
            """
            Parameterized test that confirms that the convert_to_wave method
            converts Frequency min/max to appropriate Wavelength min/max
            """
            energy_bandpass = FrequencyBandpass(min=min, max=max, unit=unit)
            wavelength_bandpass = convert_to_wave(energy_bandpass)
            assert all(
                [
                    np.isclose(wavelength_bandpass.min, min_wave),  # type:ignore
                    np.isclose(wavelength_bandpass.max, max_wave),  # type:ignore
                ]
            )



from ..core.math import find_duplicates
from .footprint import Footprint


def inner(footprints: list[Footprint], order: int = 10) -> list[int]:
    """
    Computes the inner join from a list of footprints.
    Returns the list of overlapping healpix-pixels (i.e. the duplicates)
    """
    total_pixels: list[int] = []

    for footprint in footprints:
        total_pixels.extend(footprint.query_pixels(order=order))

    return find_duplicates(total_pixels)


def outer(footprints: list[Footprint], order: int = 10) -> list[int]:
    """
    Computes the outer join from a list of footprints
    Returns the list of non-over lapping healpix pixels (i.e. the non-duplicates)
    """
    total_pixels: list[int] = []

    for footprint in footprints:
        total_pixels.extend(footprint.query_pixels(order=order))

    duplicates = find_duplicates(total_pixels)
    outer_pixels = list({x for x in total_pixels if x not in duplicates})

    return outer_pixels


def union(footprints: list[Footprint], order: int = 10) -> list[int]:
    """
    Computes the union of a list of footprints
    Returns the list of overlapping and non overlapping healpix pixels (i.e. all unique pixels)
    """
    total_pixels: list[int] = []

    for footprint in footprints:
        total_pixels.extend(footprint.query_pixels(order=order))

    return list(set(total_pixels))



import astropy.units as u  # type: ignore[import-untyped]
import numpy as np
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]

from across.tools.core.enums import TwilightType
from across.tools.ephemeris import Ephemeris, GroundEphemeris
from across.tools.ephemeris.tle_ephem import TLEEphemeris
from across.tools.visibility.constraints import DaytimeConstraint


class TestDaytimeConstraint:
    """Test suite for the DaytimeConstraint class."""

    def test_daytime_constraint_short_name(self) -> None:
        """Test that DaytimeConstraint has correct short_name."""
        constraint = DaytimeConstraint()
        assert constraint.short_name == "Daytime"

    def test_daytime_constraint_name_value(self) -> None:
        """Test that DaytimeConstraint has correct name value."""
        constraint = DaytimeConstraint()
        assert constraint.name.value == "Daytime Avoidance"

    def test_daytime_constraint_default_twilight_type(self) -> None:
        """Test that DaytimeConstraint has correct default twilight type."""
        constraint = DaytimeConstraint()
        assert constraint.twilight_type == TwilightType.ASTRONOMICAL

    def test_daytime_constraint_call_returns_ndarray(
        self, ground_ephemeris: GroundEphemeris, sky_coord: SkyCoord
    ) -> None:
        """Test that __call__ method returns numpy ndarray."""
        constraint = DaytimeConstraint()
        result = constraint(time=ground_ephemeris.timestamp, ephemeris=ground_ephemeris, coordinate=sky_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_call_returns_bool_dtype(
        self, ground_ephemeris: GroundEphemeris, sky_coord: SkyCoord
    ) -> None:
        """Test that __call__ method returns boolean dtype."""
        constraint = DaytimeConstraint()
        result = constraint(time=ground_ephemeris.timestamp, ephemeris=ground_ephemeris, coordinate=sky_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_call_returns_correct_length(
        self, ground_ephemeris: GroundEphemeris, sky_coord: SkyCoord
    ) -> None:
        """Test that __call__ method returns array of correct length."""
        constraint = DaytimeConstraint()
        result = constraint(time=ground_ephemeris.timestamp, ephemeris=ground_ephemeris, coordinate=sky_coord)
        assert len(result) == len(ground_ephemeris.timestamp)

    def test_daytime_constraint_sunset_validation_has_two_transitions(
        self, mauna_kea_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that daytime constraint identifies exactly 2 transitions (sunrise/sunset)."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        results = constraint(mauna_kea_ephemeris.timestamp, mauna_kea_ephemeris, dummy_coord)
        transitions = np.where(np.diff(results.astype(int)) != 0)[0] + 1
        assert len(transitions) == 2

    def test_daytime_constraint_sunset_validation_sunrise_time_accuracy(
        self, mauna_kea_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that daytime constraint sunrise time is close to independent calculation."""
        from astropy.coordinates import AltAz

        ephem = mauna_kea_ephemeris
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        results = constraint(ephem.timestamp, ephem, dummy_coord)
        transitions = np.where(np.diff(results.astype(int)) != 0)[0] + 1
        sunrise_idx = transitions[0]
        constraint_sunrise = ephem.timestamp[sunrise_idx]

        location = ephem.earth_location
        twilight_altitude = -6 * u.deg
        sun_altitudes = ephem.sun.transform_to(AltAz(obstime=ephem.timestamp, location=location)).alt
        below_horizon = sun_altitudes < twilight_altitude
        sunrise_mask = ~below_horizon[:-1] & below_horizon[1:]
        independent_sunrise_idx = np.where(sunrise_mask)[0]

        if len(independent_sunrise_idx) > 0:
            independent_sunrise = ephem.timestamp[independent_sunrise_idx[0]]
            time_diff_sunrise = abs((constraint_sunrise - independent_sunrise).sec) / 60
            assert time_diff_sunrise < 15

    def test_daytime_constraint_sunset_validation_sunset_time_accuracy(
        self, mauna_kea_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that daytime constraint sunset time is close to independent calculation."""
        from astropy.coordinates import AltAz

        ephem = mauna_kea_ephemeris
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        results = constraint(ephem.timestamp, ephem, dummy_coord)
        transitions = np.where(np.diff(results.astype(int)) != 0)[0] + 1
        sunset_idx = transitions[1]
        constraint_sunset = ephem.timestamp[sunset_idx]

        location = ephem.earth_location
        twilight_altitude = -6 * u.deg
        sun_altitudes = ephem.sun.transform_to(AltAz(obstime=ephem.timestamp, location=location)).alt
        below_horizon = sun_altitudes < twilight_altitude
        sunset_mask = below_horizon[:-1] & ~below_horizon[1:]
        independent_sunset_idx = np.where(sunset_mask)[0]

        if len(independent_sunset_idx) > 0:
            independent_sunset = ephem.timestamp[independent_sunset_idx[0]]
            time_diff_sunset = abs((constraint_sunset - independent_sunset).sec) / 60
            assert time_diff_sunset < 15

    def test_daytime_constraint_sunset_validation_daytime_result_is_boolean(
        self, mauna_kea_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that daytime constraint result at midday is boolean."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(mauna_kea_ephemeris.timestamp, mauna_kea_ephemeris, dummy_coord)
        midday_idx = len(mauna_kea_ephemeris.timestamp) // 2
        assert isinstance(result[midday_idx], (bool, np.bool_))

    def test_daytime_constraint_civil_twilight_type(self) -> None:
        """Test that DaytimeConstraint can be created with CIVIL twilight type."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        assert constraint.twilight_type == TwilightType.CIVIL

    def test_daytime_constraint_nautical_twilight_type(self) -> None:
        """Test that DaytimeConstraint can be created with NAUTICAL twilight type."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        assert constraint.twilight_type == TwilightType.NAUTICAL

    def test_daytime_constraint_astronomical_twilight_type(self) -> None:
        """Test that DaytimeConstraint can be created with ASTRONOMICAL twilight type."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        assert constraint.twilight_type == TwilightType.ASTRONOMICAL

    def test_daytime_constraint_sunset_twilight_type(self) -> None:
        """Test that DaytimeConstraint can be created with SUNSET twilight type."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        assert constraint.twilight_type == TwilightType.SUNSET

    def test_daytime_constraint_space_based_returns_array(
        self, test_tle_ephemeris: TLEEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based daytime constraint returns array."""
        constraint = DaytimeConstraint()
        coord = dummy_coord
        # Use the TLE ephemeris which should trigger space-based calculations
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_space_based_returns_bool_dtype(
        self, test_tle_ephemeris: TLEEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based daytime constraint returns boolean dtype."""
        constraint = DaytimeConstraint()
        coord = dummy_coord
        # Use the TLE ephemeris which should trigger space-based calculations
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_space_based_returns_correct_length(
        self, test_tle_ephemeris: TLEEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based daytime constraint returns correct length."""
        constraint = DaytimeConstraint()
        coord = dummy_coord
        # Use the TLE ephemeris which should trigger space-based calculations
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, coord)
        assert len(result) == len(test_tle_ephemeris.timestamp)

    def test_daytime_constraint_civil_twilight_ground_based_returns_ndarray(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that civil twilight ground-based result is ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_civil_twilight_ground_based_returns_bool_dtype(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that civil twilight ground-based result is bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_nautical_twilight_ground_based_returns_ndarray(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that nautical twilight ground-based result is ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_nautical_twilight_ground_based_returns_bool_dtype(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that nautical twilight ground-based result is bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_astronomical_twilight_ground_based_returns_ndarray(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that astronomical twilight ground-based result is ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_astronomical_twilight_ground_based_returns_bool_dtype(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that astronomical twilight ground-based result is bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_sunset_twilight_ground_based_returns_ndarray(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that sunset twilight ground-based result is ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_sunset_twilight_ground_based_returns_bool_dtype(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that sunset twilight ground-based result is bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        result = constraint(ground_ephemeris.timestamp, ground_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_ground_based_all_twilights_same_length(
        self, ground_ephemeris: GroundEphemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that all twilight type results have same length."""
        civil_result = DaytimeConstraint(twilight_type=TwilightType.CIVIL)(
            ground_ephemeris.timestamp, ground_ephemeris, dummy_coord
        )
        nautical_result = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)(
            ground_ephemeris.timestamp, ground_ephemeris, dummy_coord
        )
        astronomical_result = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)(
            ground_ephemeris.timestamp, ground_ephemeris, dummy_coord
        )
        sunset_result = DaytimeConstraint(twilight_type=TwilightType.SUNSET)(
            ground_ephemeris.timestamp, ground_ephemeris, dummy_coord
        )
        assert len(civil_result) == len(nautical_result) == len(astronomical_result) == len(sunset_result)

    def test_daytime_constraint_space_based_civil_twilight_returns_ndarray(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based civil twilight constraint returns ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_space_based_civil_twilight_returns_bool_dtype(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based civil twilight constraint returns bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_space_based_civil_twilight_correct_length(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based civil twilight constraint returns correct length."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.CIVIL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert len(result) == len(test_tle_ephemeris.timestamp)

    def test_daytime_constraint_space_based_nautical_twilight_returns_ndarray(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based nautical twilight constraint returns ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_space_based_nautical_twilight_returns_bool_dtype(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based nautical twilight constraint returns bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_space_based_nautical_twilight_correct_length(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based nautical twilight constraint returns correct length."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.NAUTICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert len(result) == len(test_tle_ephemeris.timestamp)

    def test_daytime_constraint_space_based_astronomical_twilight_returns_ndarray(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based astronomical twilight constraint returns ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_space_based_astronomical_twilight_returns_bool_dtype(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based astronomical twilight constraint returns bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_space_based_astronomical_twilight_correct_length(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based astronomical twilight constraint returns correct length."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.ASTRONOMICAL)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert len(result) == len(test_tle_ephemeris.timestamp)

    def test_daytime_constraint_space_based_sunset_twilight_returns_ndarray(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based sunset twilight constraint returns ndarray."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert isinstance(result, np.ndarray)

    def test_daytime_constraint_space_based_sunset_twilight_returns_bool_dtype(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based sunset twilight constraint returns bool dtype."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert result.dtype == np.bool_

    def test_daytime_constraint_space_based_sunset_twilight_correct_length(
        self, test_tle_ephemeris: Ephemeris, dummy_coord: SkyCoord
    ) -> None:
        """Test that space-based sunset twilight constraint returns correct length."""
        constraint = DaytimeConstraint(twilight_type=TwilightType.SUNSET)
        result = constraint(test_tle_ephemeris.timestamp, test_tle_ephemeris, dummy_coord)
        assert len(result) == len(test_tle_ephemeris.timestamp)



from pydantic import Field

from .base import BaseSchema


class RollAngle(BaseSchema):
    """
    Class to represent and validate a roll angle
        constraint: Must be (-360.0 >= a >= 360.0)
    """

    value: float = Field(ge=-360, le=360)



from enum import Enum


class FrequencyUnit(str, Enum):
    """
    Enum to represent the bandpass frequency types
    """

    Hz = "Hz"
    kHz = "kHz"  # noqa: N815
    MHz = "MHz"
    GHz = "GHz"
    THz = "THz"



from collections.abc import Generator
from datetime import datetime
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from across.tools.tle.tle import TLEFetch


@pytest.fixture
def mock_spacetrack() -> Generator[MagicMock]:
    """Return a mock SpaceTrackClient."""
    with patch("across.tools.tle.tle.SpaceTrackClient") as mock:
        yield mock


@pytest.fixture
def valid_spacetrack_tle_response() -> str:
    """Return a valid TLE response."""
    return (
        "1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"
    )


@pytest.fixture
def multi_norad_spacetrack_tle_response() -> str:
    """Return a TLE response containing multiple NORAD IDs and repeated entries."""
    return (
        "1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537\n"
        "1 28485U 05055A   08263.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 28485  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537\n"
        "1 25544U 98067A   08262.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"
    )


@pytest.fixture
def duplicate_norad_spacetrack_tle_response() -> str:
    """Return a TLE response with multiple entries for the same NORAD ID at different epochs."""
    return (
        "1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537\n"
        "1 25544U 98067A   08263.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537\n"
        "1 25544U 98067A   08262.51782528 -.00002182  00000-0 -11606-4 0  2927\n"
        "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"
    )


@pytest.fixture
def empty_spacetrack_tle_response() -> str:
    """Return an empty TLE response."""
    return ""


@pytest.fixture
def mock_spacetrack_instance(mock_spacetrack: MagicMock) -> MagicMock:
    """Mocked spacetrack instance"""
    mock_instance = MagicMock()
    mock_instance.gp = MagicMock()

    mock_spacetrack.return_value.__enter__.return_value = mock_instance

    return mock_instance


@pytest.fixture
def valid_tle_data(valid_spacetrack_tle_response: str) -> dict[str, Any]:
    """Fixture providing valid TLE data."""
    return {
        "norad_id": 25544,
        "satellite_name": "ISS (ZARYA)",
        "tle1": valid_spacetrack_tle_response.split("\n")[0],
        "tle2": valid_spacetrack_tle_response.split("\n")[1],
    }


@pytest.fixture
def tle_fetch_object() -> Generator[TLEFetch]:
    """Example TLEFetch object."""
    yield TLEFetch(
        satellites=[{"name": "ISS", "id": 25544}],
        epoch=datetime(2008, 9, 20),
        spacetrack_user="user",
        spacetrack_pwd="pass",
    )



from typing import Any

import pytest

from across.tools.core.schemas.coordinate import Coordinate
from across.tools.core.schemas.polygon import Polygon


class TestPolygonSchema:
    """Test suite for the Polygon schema."""

    def test_valid_polygon(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test Polygon with valid data."""
        polygon = Polygon(**valid_polygon_data)
        assert len(polygon.coordinates) == 4

    def test_empty_coordinates(self) -> None:
        """Test Polygon with empty coordinates list."""
        with pytest.raises(ValueError, match="Invalid polygon, coordinates cannot be empty."):
            Polygon(coordinates=[])

    def test_auto_close_polygon(self) -> None:
        """Test that polygon auto-closes when first and last points don't match."""
        coords = [Coordinate(ra=0, dec=0), Coordinate(ra=1, dec=1), Coordinate(ra=1, dec=0)]
        polygon = Polygon(coordinates=coords)
        assert polygon.coordinates[0] == polygon.coordinates[-1]
        assert len(polygon.coordinates) == 4

    def test_too_few_coordinates(self) -> None:
        """Test Polygon with less than 3 unique coordinates."""
        coords = [Coordinate(ra=0, dec=0), Coordinate(ra=1, dec=1), Coordinate(ra=0, dec=0)]
        with pytest.raises(ValueError, match="Invalid polygon, must contain more than 3 unique coordinates"):
            Polygon(coordinates=coords)

    def test_polygon_equality(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test polygon equality comparison."""
        polygon1 = Polygon(**valid_polygon_data)
        polygon2 = Polygon(**valid_polygon_data)
        assert polygon1 == polygon2

    def test_polygon_inequality(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test polygon inequality comparison."""
        polygon1 = Polygon(**valid_polygon_data)
        different_coords = [
            Coordinate(ra=0, dec=0),
            Coordinate(ra=2, dec=2),
            Coordinate(ra=2, dec=0),
            Coordinate(ra=0, dec=0),
        ]
        polygon2 = Polygon(coordinates=different_coords)
        assert polygon1 != polygon2

    def test_polygon_repr(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test polygon string representation."""
        polygon = Polygon(**valid_polygon_data)
        repr_str = repr(polygon)
        assert repr_str.startswith("Polygon(")
        assert repr_str.endswith(")")
        assert "Coordinate" in repr_str

    def test_polygon_eq_non_polygon(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test polygon equality with non-polygon object."""
        polygon = Polygon(**valid_polygon_data)
        assert (polygon == "not a polygon") is False

    def test_polygon_eq_different_length(self, valid_polygon_data: dict[str, Any]) -> None:
        """Test polygon equality with different length coordinates."""
        polygon1 = Polygon(**valid_polygon_data)
        coords = [
            Coordinate(ra=0, dec=0),
            Coordinate(ra=1, dec=1),
            Coordinate(ra=1, dec=0),
            Coordinate(ra=0.5, dec=0.5),
            Coordinate(ra=0, dec=0),
        ]
        polygon2 = Polygon(coordinates=coords)
        assert polygon1 != polygon2



from datetime import datetime, timedelta

import astropy.units as u  # type: ignore[import-untyped]
import numpy as np
import spiceypy as spice  # type: ignore[import-untyped]
from astropy.coordinates import (  # type: ignore[import-untyped]
    GCRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]
from astropy.utils.data import download_file  # type: ignore[import-untyped]

from .base import Ephemeris

NAIF_LEAP_SECONDS_URL = "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls"
NAIF_PLANETARY_EPHEMERIS_URL = "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/de442s.bsp"
NAIF_EARTH_ORIENTATION_PARAMETERS_URL = (
    "https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/earth_latest_high_prec.bpc"
)


class SPICEEphemeris(Ephemeris):
    """
    SPICEEphemeris class for calculating spacecraft ephemeris using SPICE
    (Spacecraft Planet Instrument C-matrix Events) kernels.

    This class extends the base Ephemeris class to provide specialized
    functionality for computing spacecraft trajectories using  kernels.

    Parameters
    ----------
    begin : datetime or Time
        Start time of the ephemeris calculation
    end : datetime or Time
        End time of the ephemeris calculation
    step_size : int or TimeDelta or timedelta, default=60
        Time step between ephemeris points in seconds
    spice_kernel_url : str, optional
        URL to download the spacecraft SPICE kernel
    naif_id : int, optional
        Navigation and Ancillary Information Facility (NAIF) ID of the
        spacecraft/object in JPL Horizons or SPICE kernel

    Attributes
    ----------
    naif_id : int
        NAIF ID of object for JPL Horizons or SPICE Kernel
    spice_kernel_url : str
        URL of spacecraft SPICE Kernel
    gcrs : SkyCoord
        Calculated positions in GCRS frame
    earth_location : EarthLocation
        Calculated positions in ITRS frame

    Methods
    -------
    prepare_data()
        Loads necessary SPICE kernels and calculates spacecraft trajectory

    Notes
    -----
    Requires SPICE kernels for:
    - Leap seconds
    - Planetary ephemeris
    - Earth orientation parameters
    - Spacecraft trajectory
    The class automatically handles downloading and caching of required SPICE
    kernels.
    """

    # NAIF ID of object for JPL Horizons or Spice Kernel
    naif_id: int | None = None

    # URL of spacecraft SPICE Kernel
    spice_kernel_url: str | None = None

    def __init__(
        self,
        begin: datetime | Time,
        end: datetime | Time,
        step_size: int | TimeDelta | timedelta = 60,
        spice_kernel_url: str | None = None,
        naif_id: int | None = None,
    ) -> None:
        super().__init__(begin, end, step_size)
        self.spice_kernel_url = spice_kernel_url
        self.naif_id = naif_id

    def prepare_data(self) -> None:
        """Loading SPICE kernels to calculate SPICE based ephemeris."""
        # Check that parameters needed for SPICE are set
        if self.spice_kernel_url is None:
            raise ValueError("No SPICE kernel URL provided")
        if self.naif_id is None:
            raise ValueError("No NAIF ID provided")

        # Download and load SPICE kernels if not already loaded
        kernel_urls = [
            NAIF_LEAP_SECONDS_URL,
            NAIF_PLANETARY_EPHEMERIS_URL,
            NAIF_EARTH_ORIENTATION_PARAMETERS_URL,
            self.spice_kernel_url,
        ]
        loaded_kernels = [str(spice.kdata(i, "all")[0]) for i in range(spice.ktotal("all"))]

        for url in kernel_urls:
            kernel_file = download_file(url, cache=True)
            if kernel_file not in loaded_kernels:
                spice.furnsh(kernel_file)

        # Generate array of times for ephemeris calculation in Ephemeris time (ET) format
        start_et = spice.str2et(str(self.begin.datetime))
        end_et = start_et + self.step_size.to_value(u.s) * len(self.timestamp)
        time_intervals = np.arange(start_et, end_et, self.step_size.to_value(u.s))

        # Get position and velocity vectors (states) in GCRS/J2000 frame
        states = np.array(
            [spice.spkezr(str(self.naif_id), et, "J2000", "NONE", "399")[0] for et in time_intervals]
        )

        # Create GCRS SkyCoord from states array
        self.gcrs = SkyCoord(
            CartesianRepresentation(states[:, :3].T * u.km).with_differentials(
                CartesianDifferential(states[:, 3:].T * u.km / u.s)
            ),
            frame=GCRS(obstime=self.timestamp),
        )

        # Transform to ITRS and get Earth location
        self.earth_location = self.gcrs.transform_to("itrs").earth_location


def compute_spice_ephemeris(
    begin: datetime | Time,
    end: datetime | Time,
    step_size: int | timedelta | TimeDelta,
    spice_kernel_url: str,
    naif_id: int,
) -> Ephemeris:
    """
    Compute space ephemeris data using SPICE kernels.

    Parameters
    ----------
    begin : Union[datetime, Time]
        Start date and time for ephemeris computation
    end : Union[datetime, Time]
        End date and time for ephemeris computation
    step_size : Union[int, timedelta, TimeDelta]
        Time step size between ephemeris points, in seconds if int
    spice_kernel_url : str
        URL to the SPICE kernel file
    naif_id : int
        Navigation and Ancillary Information Facility (NAIF) object identifier
        (e.g., 301 for Moon, -48 for HST)

    Returns
    -------
    Ephemeris
        An Ephemeris object containing the computed ephemeris data

    Notes
    -----
    This function uses the SPICE system to compute high-precision
    ephemeris data for celestial bodies in space-based reference frame.
    Examples
    --------
    >>> from datetime import datetime
    >>> begin = datetime(2023, 1, 1)
    >>> end = datetime(2023, 1, 2)
    >>> moon_ephemeris = compute_spice_ephemeris(begin, end, 60, 'https://path/to/spice_kernel.bsp', 301)
    """
    # Compute the ephemeris using the SPICEEphemeris class
    ephemeris = SPICEEphemeris(
        naif_id=naif_id, begin=begin, end=end, step_size=step_size, spice_kernel_url=spice_kernel_url
    )
    ephemeris.compute()
    return ephemeris



from datetime import datetime

import pytest

from across.tools.ephemeris.tle_ephem import TLEEphemeris


class TestTLEEphemeris:
    """Test suite for the TLEEphemeris class."""

    def test_tle_ephemeris_no_naif_id(self) -> None:
        """Test that TLEEphemeris raises ValueError when no NAIF ID is provided."""
        begin = datetime(2023, 1, 1)
        end = datetime(2023, 1, 2)
        with pytest.raises(ValueError, match="No TLE provided"):
            ephemeris = TLEEphemeris(begin=begin, end=end)
            ephemeris.prepare_data()



from datetime import datetime, timedelta

import astropy.units as u  # type: ignore[import-untyped]
import astroquery.jplhorizons as jpl  # type: ignore[import-untyped]
from astropy.coordinates import (  # type: ignore[import-untyped]
    GCRS,
    ITRS,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]

from .base import Ephemeris


class JPLEphemeris(Ephemeris):
    """
    JPL Horizons-based ephemeris calculation.

    This class provides functionality to calculate ephemeris data using JPL
    Horizons system. It requires a Navigation and Ancillary Information
    Facility (NAIF) ID to identify the celestial object of interest.

    Parameters
    ----------
    begin : datetime or Time
        Start time of the ephemeris calculation
    end : datetime or Time
        End time of the ephemeris calculation
    step_size : int or TimeDelta or timedelta, default=60
        Time step between ephemeris points in seconds
    naif_id : int, optional
        NAIF ID of the celestial object for JPL Horizons query

    Attributes
    ----------
    naif_id : int
        NAIF ID of object for JPL Horizons or Spice Kernel
    gcrs : SkyCoord
        Geocentric celestial reference system coordinates
    earth_location : EarthLocation
        Earth-fixed coordinates of the object

    Methods
    -------
    prepare_data()
        Calculate ephemeris based on JPL Horizons data

    Notes
    -----
    The class uses the JPL Horizons API to fetch vector data for the specified
    celestial object and converts it to both GCRS and ITRS coordinate frames.

    Raises
    ------
    ValueError
        If no NAIF ID is provided when preparing the data
    """

    # NAIF ID of object for JPL Horizons or Spice Kernel
    naif_id: int | None = None

    def __init__(
        self,
        begin: datetime | Time,
        end: datetime | Time,
        step_size: int | TimeDelta | timedelta = 60,
        naif_id: int | None = None,
    ) -> None:
        super().__init__(begin, end, step_size)
        self.naif_id = naif_id

    def prepare_data(self) -> None:
        """Calculate ephemeris based on JPL Horizons data."""
        # Check that parameters needed for JPL Horizons are set
        if self.naif_id is None:
            raise ValueError("No NAIF ID provided")

        # Calculate the number of steps between start and stop
        num_steps = len(self.timestamp) - 1

        # Create a time range dictionary for Horizons
        horizons_range = {
            "start": str(self.begin.tdb.datetime),
            "stop": str(self.end.tdb.datetime),
            "step": str(num_steps),
        }

        # Fetch the ephemeris vector data from Horizons
        horizons_ephemeris = jpl.Horizons(
            id=self.naif_id,
            location="500@399",
            epochs=horizons_range,
            id_type=None,
        )
        horizons_vectors = horizons_ephemeris.vectors(refplane="earth")

        # Create a GCRS SkyCoord object from the ephemeris data
        self.gcrs = SkyCoord(
            CartesianRepresentation(
                horizons_vectors["x"].to(u.km),
                horizons_vectors["y"].to(u.km),
                horizons_vectors["z"].to(u.km),
            ).with_differentials(
                CartesianDifferential(
                    horizons_vectors["vx"],
                    horizons_vectors["vy"],
                    horizons_vectors["vz"],
                )
            ),
            frame=GCRS(obstime=self.timestamp),
        )

        # Calculate the ITRS coordinates and Earth Location
        itrs = self.gcrs.transform_to(ITRS(obstime=self.timestamp))
        self.earth_location = itrs.earth_location


def compute_jpl_ephemeris(
    begin: datetime | Time,
    end: datetime | Time,
    step_size: int | timedelta | TimeDelta,
    naif_id: int,
) -> Ephemeris:
    """
    Compute space ephemeris data using JPL Horizons system.

    Parameters
    ----------
    begin : Union[datetime, Time]
        Start date and time for ephemeris computation
    end : Union[datetime, Time]
        End date and time for ephemeris computation
    step_size : Union[int, timedelta, TimeDelta]
        Time step size between ephemeris points, in seconds if int
    naif_id : int
        Navigation and Ancillary Information Facility (NAIF) object identifier
        (e.g., 301 for Moon. -48 for HST)

    Returns
    -------
    Ephemeris
        An Ephemeris object containing the computed ephemeris data

    Notes
    -----
    This function uses the JPL Horizons system to compute high-precision
    ephemeris data for celestial bodies in space-based reference frame.
    Examples
    --------
    >>> from datetime import datetime
    >>> begin = datetime(2023, 1, 1)
    >>> end = datetime(2023, 1, 2)
    >>> moon_ephemeris = compute_jpl_ephemeris(begin, end, 60, 301)
    """
    # Compute the ephemeris using the JPLEphemeris class
    ephemeris = JPLEphemeris(naif_id=naif_id, begin=begin, end=end, step_size=step_size)
    ephemeris.compute()
    return ephemeris



from typing import Any, Generic, TypeVar
from uuid import UUID

import numpy as np
import plotly.graph_objects as go
from pydantic import Field, model_validator

from across.tools.core.plotting import plot_joint_visibility_windows, plot_visibility_windows

from ..core.enums import ConstraintType
from ..core.schemas import (
    AstropyDateTime,
    AstropyTimeDelta,
)
from .base import Visibility

T = TypeVar("T", bound=Visibility)


class JointVisibility(Visibility, Generic[T]):
    """
    Computes joint visibility windows between multiple instruments.

    This class takes a list of Visibility objects with identical timestamp grids
    and computes the intersection of their visibility periods.

    Parameters
    ----------
    visibilities : list[T]
        List of Visibility or Visibility child objects with identical timestamp grids.
    instrument_ids : list[UUID]
        List of IDs of the instruments belonging to the Visibility objects.
    """

    # Parameters
    visibilities: list[T] = Field(default_factory=list, exclude=True)
    instrument_ids: list[UUID] = Field(default_factory=list, exclude=True)

    # Values derived from input parameters
    step_size: AstropyTimeDelta = Field(default=None)
    begin: AstropyDateTime = Field(default=None)
    end: AstropyDateTime = Field(default=None)
    observatory_name: str = Field(default="", exclude=True)

    @model_validator(mode="before")
    @classmethod
    def validate_parameters(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Validate and synchronize coordinate, begin/end, and step size values.
        This method ensures that all the input Visibility objects have the same
        coordinates, begin and end times, and step sizes. If any of these values
        differ between Visibilities, a ValueError is raised.
        When these equalities have been validated, this method runs the base Visibility
        validate_parameters method to validate other parameter values.
        """
        # Check that all ra/dec values are the same within tolerance (~15 arcsec)
        tolerance = 15.0 / 3600.0

        ras = np.array([(visibility.ra) for visibility in values["visibilities"]], dtype=float)
        decs = np.array([(visibility.dec) for visibility in values["visibilities"]], dtype=float)

        if not (np.allclose(ras, ras[0], atol=tolerance) and np.allclose(decs, decs[0], atol=tolerance)):
            raise ValueError(
                f"All input visibilities must have the same coordinate within {tolerance} degrees"
            )
        values["ra"] = values["visibilities"][0].ra
        values["dec"] = values["visibilities"][0].dec

        # Check that all begin/end values are the same
        if (
            not len(set([visibility.begin for visibility in values["visibilities"]])) == 1
            or not len(set([visibility.end for visibility in values["visibilities"]])) == 1
        ):
            raise ValueError("All begin and end times must be the same")
        values["begin"] = values["visibilities"][0].begin
        values["end"] = values["visibilities"][0].end

        # Check that all step sizes are the same
        if not len(set([visibility.step_size for visibility in values["visibilities"]])) == 1:
            raise ValueError("All step sizes must be the same")
        values["step_size"] = values["visibilities"][0].step_size

        return values

    def _constraint(self, i: int) -> ConstraintType:
        """
        For a given index, return the constraint
        from the first visibility that is actually constrained.
        """
        # Safely handle out-of-bounds indices
        if i < 0 or i >= len(self.timestamp) if self.timestamp else True:
            return ConstraintType.WINDOW

        # Find the first visibility that is constrained at this index
        for vis in self.visibilities:
            if vis.inconstraint[i]:
                return vis._constraint(i)

        return ConstraintType.UNKNOWN

    def _get_id(self, i: int) -> UUID:
        """
        For a given index, find the observatory ID of the first visibility that is constrained.
        """
        # Safely handle out-of-bounds indices
        if self.timestamp is None or i < 0 or i >= len(self.timestamp):
            return self.visibilities[0]._get_id(i) if self.visibilities else UUID(int=0)

        for vis in self.visibilities:
            if vis.inconstraint[i]:
                return vis._get_id(i)

        # Unknown constraint found, so just return the first visibility's observatory ID
        return self.visibilities[0]._get_id(i) if self.visibilities else UUID(int=0)

    def _get_name(self, i: int) -> str:
        """
        For a given index, get the name of the first instrument that is constrained.
        """
        # Safely handle out-of-bounds indices by using first visibility's name
        if self.timestamp is None or i < len(self.timestamp):
            for vis in self.visibilities:
                if i >= 0 and i < len(vis.inconstraint) and vis.inconstraint[i]:
                    return vis.observatory_name

        # Return first visibility's name as fallback
        if self.visibilities:
            return self.visibilities[0].observatory_name
        return ""

    def prepare_data(self) -> None:
        """
        Compute joint visibility by ANDing all inconstraint arrays.

        Raises
        ------
        ValueError
            If visibilities list is empty or if visibilities have different timestamp grids.
        """
        if not self.visibilities:
            raise ValueError("No visibilities provided for joint visibility calculation")

        # Compute joint visibility by ORing all inconstraint arrays
        self.inconstraint = np.any([vis.inconstraint for vis in self.visibilities], axis=0)

    def _merge_computed_values(self) -> None:
        """
        Abstract method to merge computed values from constraints.
        """
        for visibility in self.visibilities:
            if visibility.computed_values is not None:
                self.computed_values.merge(visibility.computed_values)

    def plot(
        self, fig: go.Figure | None = None, offset: int | float = 0, width: int = 700, height: int = 1000
    ) -> go.Figure:
        """
        Method to visualize joint visibility windows using plotly.
        Plots the individual instrument visibility windows and the regions
        of joint visibility on one figure. Calls the across-tools plotting
        core functionality and configures the plot layout to user specifications.

        Parameters
        ----------
        fig : go.Figure, optional
            An existing plotly figure to add to, by default None
        offset : int | float, optional
            The x-axis offset to plot new visibility windows, by default 0
        width: int, optional
            The width of the plot, in pixels. Defaults to 700.
        height: int, optional
            The height of the plot, in pixels. Defaults to 1000.

        Returns
        -------
        go.Figure
            The plotly figure containing the footprint plot
        """
        if fig is None:
            fig = go.Figure()
        tickvals = []
        ticktext = []
        for i, visibility in enumerate(self.visibilities):
            fig = plot_visibility_windows(
                visibility_windows=[window.model_dump() for window in visibility.visibility_windows],
                observatory_name=visibility.observatory_name,
                fig=fig,
                offset=offset + i + 1,
            )
            tickvals.append(offset + i + 1)
            ticktext.append(visibility.observatory_name)

        min_extent = min(tickvals)
        max_extent = max(tickvals)
        fig = plot_joint_visibility_windows(
            visibility_windows=[window.model_dump() for window in self.visibility_windows],
            min_extent=min_extent,
            max_extent=max_extent,
            fig=fig,
        )

        fig.update_layout(
            title="Visibility Windows",
            yaxis=dict(
                title="Time (UTC)",
                range=[self.end.to_datetime(), self.begin.to_datetime()],  # descending time
                type="date",
                autorange=False,  # don't resize
            ),
            xaxis=dict(
                title="Visibility Windows",
                tickvals=tickvals,
                ticktext=ticktext,
            ),
            width=width,
            height=height,
        )
        return fig


def compute_joint_visibility(
    visibilities: list[T],
    instrument_ids: list[UUID],
    min_vis: int = 0,
) -> JointVisibility[T]:
    """
    Compute joint visibility windows for any number of instrument Visibilities.
    Assumes that the visibilities are in the same order as instrument_ids.

    Parameters
    ----------
    visibilities: list[T]
        List of Visibility objects or children objects of Visibility.
    instrument_ids: list[UUID]
        List of IDs of the instruments belonging to the Visibility objects.
    min_vis: int, optional
        Minimum visibility duration in seconds for a window to be included.
        Default is 0.

    Returns
    -------
    list[VisibilityWindow]
        List of VisibilityWindows capturing joint visibility between all inputted instruments.
    """
    joint_vis = JointVisibility(
        visibilities=visibilities,
        instrument_ids=instrument_ids,
        min_vis=min_vis,
    )
    joint_vis.compute()
    return joint_vis



from datetime import datetime
from typing import Any

import pytest
from pydantic import ValidationError

from across.tools.core.schemas.tle import TLE, TLEBase


class TestTLESchema:
    """Test suite for the TLE schema."""

    def test_tle_base_norad_id(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase norad_id with valid data."""
        tle_base = TLEBase(**valid_tle_data)
        assert tle_base.norad_id == 25544

    def test_tle_base_satellite_name(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase satellite_name with valid data."""
        tle_base = TLEBase(**valid_tle_data)
        assert tle_base.satellite_name == "ISS (ZARYA)"

    def test_tle_base_tle1(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase tle1 with valid data."""
        tle_base = TLEBase(**valid_tle_data)
        assert tle_base.tle1 == "1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927"

    def test_tle_base_tle2(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase tle2 with valid data."""
        tle_base = TLEBase(**valid_tle_data)
        assert tle_base.tle2 == "2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537"

    def test_tle_base_norad_id_optional(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase with optional fields set to None."""
        tle_base = TLEBase(tle1=valid_tle_data["tle1"], tle2=valid_tle_data["tle2"])
        assert tle_base.norad_id is None

    def test_tle_base_satellite_name_optional(self, valid_tle_data: dict[str, Any]) -> None:
        """Test TLEBase with optional fields set to None."""
        tle_base = TLEBase(tle1=valid_tle_data["tle1"], tle2=valid_tle_data["tle2"])
        assert tle_base.satellite_name is None

    def test_tle_base_invalid_tle1_length(self) -> None:
        """Test TLEBase with invalid tle1 length."""
        with pytest.raises(ValueError):
            TLEBase(
                tle1="12345",
                tle2="2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537",
            )

    def test_tle_base_invalid_tle2_length(self) -> None:
        """Test TLEBase with invalid tle2 length."""
        with pytest.raises(ValueError):
            TLEBase(
                tle1="1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927",
                tle2="12345",
            )

    def test_tle_epoch_year_calculation(self, valid_tle_data: dict[str, Any]) -> None:
        """Test the epoch calculation of the TLE class."""
        tle = TLE(**valid_tle_data)
        assert tle.epoch.year == 2008

    def test_tle_epoch_month_calculation(self, valid_tle_data: dict[str, Any]) -> None:
        """Test the epoch calculation of the TLE class."""
        tle = TLE(**valid_tle_data)
        assert tle.epoch.month == 9

    def test_tle_epoch_day_calculation(self, valid_tle_data: dict[str, Any]) -> None:
        """Test the epoch calculation of the TLE class."""
        tle = TLE(**valid_tle_data)
        assert tle.epoch.day == 20

    def test_tle_epoch_day_matches_calculated_value(self, valid_tle_data: dict[str, Any]) -> None:
        """Test the epoch calculation of the TLE class."""
        with pytest.raises(ValidationError):
            TLE(**valid_tle_data, epoch=datetime(2008, 9, 21))

    def test_tle_line_one_doesnt_start_with_one(self, valid_tle_data: dict[str, Any]) -> None:
        """Test that TLE line 1 must start with '1'."""
        with pytest.raises(ValidationError):
            TLE(
                tle2=valid_tle_data["tle2"],
                tle1="2 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927",
            )

    def test_tle_line_two_doesnt_start_with_two(self, valid_tle_data: dict[str, Any]) -> None:
        """Test that TLE line 2 must start with '2'."""
        with pytest.raises(ValidationError):
            TLE(
                tle1=valid_tle_data["tle1"],
                tle2="1 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537",
            )



from pydantic import BaseModel, ConfigDict


class BaseSchema(BaseModel):
    """
    Base class for schemas.

    This class provides a base implementation for schemas and defines the `from_attributes` method.
    Subclasses can inherit from this class and override the `from_attributes` method to define their
        own schema logic.
    """

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))



from typing import Any, cast

import numpy as np
import numpy.typing as npt
import pytest
from astropy import units as u  # type: ignore[import-untyped]
from astropy.coordinates import AltAz, EarthLocation, SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.core.schemas.base import BaseSchema
from across.tools.core.schemas.coordinate import Coordinate
from across.tools.core.schemas.custom_types import (
    AstropyAltAz,
    AstropyAngles,
    AstropyDateTime,
    AstropySkyCoords,
    AstropyTimeDelta,
    NumpyArray,
)
from across.tools.core.schemas.visibility import VisibilityComputedValues


class DummyModel(BaseSchema):
    """Test model for BaseSchema."""

    pass


class DummyModelTwo(BaseSchema):
    """Test model for BaseSchema."""

    pass


# Concrete model classes for custom type tests
class ModelWithTimestamp(BaseSchema):
    """Model with AstropyDateTime field."""

    timestamp: AstropyDateTime


class ModelWithTimestamps(BaseSchema):
    """Model with multiple timestamps."""

    timestamps: AstropyDateTime


class ModelWithDuration(BaseSchema):
    """Model with AstropyTimeDelta field."""

    duration: AstropyTimeDelta


class ModelWithAltitude(BaseSchema):
    """Model with single AstropyAngles field."""

    altitude: AstropyAngles


class ModelWithAltitudes(BaseSchema):
    """Model with multiple AstropyAngles fields."""

    altitudes: AstropyAngles


class ModelWithCoordinate(BaseSchema):
    """Model with single AstropySkyCoords field."""

    coordinate: AstropySkyCoords


class ModelWithCoordinates(BaseSchema):
    """Model with multiple AstropySkyCoords fields."""

    coordinates: AstropySkyCoords


class ModelWithAltAz(BaseSchema):
    """Model with single AstropyAltAz field."""

    coordinate: AstropyAltAz


class ModelWithAltAzs(BaseSchema):
    """Model with multiple AstropyAltAz fields."""

    coordinates: AstropyAltAz


class ModelWithValues(BaseSchema):
    """Model with NumpyArray field."""

    values: NumpyArray


class ModelWithValue(BaseSchema):
    """Model with single NumpyArray value field."""

    value: NumpyArray


class ModelWithMatrix(BaseSchema):
    """Model with 2D NumpyArray matrix field."""

    matrix: NumpyArray


class ModelWithData(BaseSchema):
    """Model with dict of NumpyArray data field."""

    data: dict[str, NumpyArray]


class ModelWithOptionalValues(BaseSchema):
    """Model with optional NumpyArray values field."""

    values: NumpyArray | None = None


class ComprehensiveModel(BaseSchema):
    """Model with multiple custom types for integration testing."""

    timestamp: AstropyDateTime
    angles: AstropyAngles
    coordinates: AstropySkyCoords
    values: NumpyArray


class OptionalModel(BaseSchema):
    """Model with optional custom type fields."""

    angle: AstropyAngles | None = None
    coordinate: AstropySkyCoords | None = None
    values: NumpyArray | None = None


class AngleModel(BaseSchema):
    """Model with AstropyAngles field for error testing."""

    angle: AstropyAngles


class TimeModel(BaseSchema):
    """Model with AstropyDateTime field for error testing."""

    time: AstropyDateTime


class CoordModel(BaseSchema):
    """Model with AstropySkyCoords field for error testing."""

    coord: AstropySkyCoords


class AltAzModel(BaseSchema):
    """Model with AstropyAltAz field for error testing."""

    coord: AstropyAltAz


class MultiModel(BaseSchema):
    """Model with multiple custom type validators for independent error testing."""

    angle: AstropyAngles
    time: AstropyDateTime
    coord: AstropySkyCoords
    altaz: AstropyAltAz | None = None


@pytest.fixture
def test_model_class() -> type[DummyModel]:
    """Return a DummyModel class."""
    return DummyModel


@pytest.fixture
def test_model() -> BaseSchema:
    """Return a DummyModel instance."""
    return DummyModel()


@pytest.fixture
def test_model_two() -> BaseSchema:
    """Return a DummyModel instance."""
    return DummyModelTwo()


@pytest.fixture
def model_with_timestamp() -> type[ModelWithTimestamp]:
    """Return ModelWithTimestamp class."""
    return ModelWithTimestamp


@pytest.fixture
def model_with_timestamps() -> type[ModelWithTimestamps]:
    """Return ModelWithTimestamps class."""
    return ModelWithTimestamps


@pytest.fixture
def model_with_duration() -> type[ModelWithDuration]:
    """Return ModelWithDuration class."""
    return ModelWithDuration


@pytest.fixture
def model_with_altitude() -> type[ModelWithAltitude]:
    """Return ModelWithAltitude class."""
    return ModelWithAltitude


@pytest.fixture
def model_with_altitudes() -> type[ModelWithAltitudes]:
    """Return ModelWithAltitudes class."""
    return ModelWithAltitudes


@pytest.fixture
def model_with_coordinate() -> type[ModelWithCoordinate]:
    """Return ModelWithCoordinate class."""
    return ModelWithCoordinate


@pytest.fixture
def model_with_coordinates() -> type[ModelWithCoordinates]:
    """Return ModelWithCoordinates class."""
    return ModelWithCoordinates


@pytest.fixture
def model_with_altaz() -> type[ModelWithAltAz]:
    """Return ModelWithAltAz class."""
    return ModelWithAltAz


@pytest.fixture
def model_with_altazs() -> type[ModelWithAltAzs]:
    """Return ModelWithAltAzs class."""
    return ModelWithAltAzs


@pytest.fixture
def model_with_values() -> type[ModelWithValues]:
    """Return ModelWithValues class."""
    return ModelWithValues


@pytest.fixture
def model_with_value() -> type[ModelWithValue]:
    """Return ModelWithValue class."""
    return ModelWithValue


@pytest.fixture
def model_with_matrix() -> type[ModelWithMatrix]:
    """Return ModelWithMatrix class."""
    return ModelWithMatrix


@pytest.fixture
def model_with_data() -> type[ModelWithData]:
    """Return ModelWithData class."""
    return ModelWithData


@pytest.fixture
def model_with_optional_values() -> type[ModelWithOptionalValues]:
    """Return ModelWithOptionalValues class."""
    return ModelWithOptionalValues


@pytest.fixture
def comprehensive_model() -> type[ComprehensiveModel]:
    """Return ComprehensiveModel class."""
    return ComprehensiveModel


@pytest.fixture
def optional_model() -> type[OptionalModel]:
    """Return OptionalModel class."""
    return OptionalModel


@pytest.fixture
def angle_model() -> type[AngleModel]:
    """Return AngleModel class."""
    return AngleModel


@pytest.fixture
def time_model() -> type[TimeModel]:
    """Return TimeModel class."""
    return TimeModel


@pytest.fixture
def coord_model() -> type[CoordModel]:
    """Return CoordModel class."""
    return CoordModel


@pytest.fixture
def altaz_model() -> type[AltAzModel]:
    """Return AltAzModel class."""
    return AltAzModel


@pytest.fixture
def multi_model() -> type[MultiModel]:
    """Return MultiModel class."""
    return MultiModel


@pytest.fixture
def valid_coordinates() -> list[Coordinate]:
    """Return a list of valid coordinates."""
    return [
        Coordinate(ra=0, dec=0),
        Coordinate(ra=1, dec=1),
        Coordinate(ra=1, dec=0),
        Coordinate(ra=0, dec=0),
    ]


@pytest.fixture
def valid_polygon_data(valid_coordinates: list[Coordinate]) -> dict[str, Any]:
    """Return a dictionary containing valid polygon data."""
    return {"coordinates": valid_coordinates}


@pytest.fixture
def coord_standard() -> Coordinate:
    """Fixture for a standard coordinate with RA=10.0, Dec=20.0."""
    return Coordinate(ra=10.0, dec=20.0)


@pytest.fixture
def coord_small_values() -> Coordinate:
    """Fixture for a coordinate with very small values."""
    return Coordinate(ra=0.0000123, dec=0.0000456)


@pytest.fixture
def coord_large_values() -> Coordinate:
    """Fixture for a coordinate with extreme values."""
    return Coordinate(ra=123.456789, dec=45.678901)


@pytest.fixture
def coord_negative_ra() -> Coordinate:
    """Fixture for a coordinate with negative RA."""
    return Coordinate(ra=-45.678901, dec=0.0)


@pytest.fixture
def coord_extreme() -> Coordinate:
    """Fixture for a coordinate with extreme values."""
    return Coordinate(ra=360, dec=90)


@pytest.fixture
def coord_extreme_negative() -> Coordinate:
    """Fixture for a coordinate with extreme negative values."""
    return Coordinate(ra=-360, dec=-90)


@pytest.fixture
def altaz_frame() -> AltAz:
    """Create a standard AltAz frame for testing."""
    location = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    time = Time("2020-01-01 12:00:00")
    return AltAz(location=location, obstime=time)


@pytest.fixture
def standard_time() -> Time:
    """Create a standard Time object for testing."""
    return Time("2020-01-01 12:00:00")


@pytest.fixture
def reference_datetime_string() -> str:
    """Reference datetime string for testing."""
    return "2020-01-01T12:00:00"


@pytest.fixture
def single_angle() -> Any:
    """Single angle quantity for testing."""
    return 45.0 * u.deg


@pytest.fixture
def angle_array() -> Any:
    """Array of angles for testing."""
    return [30.0, 45.0, 60.0] * u.deg


@pytest.fixture
def numpy_array_1d() -> npt.NDArray[np.float64]:
    """1D numpy array for testing."""

    return np.array([1.0, 2.0, 3.0], dtype=np.float64)


@pytest.fixture
def numpy_array_2d() -> npt.NDArray[np.float64]:
    """2D numpy array for testing."""

    return np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64)


@pytest.fixture
def numpy_dict_arrays() -> dict[str, npt.NDArray[np.float64]]:
    """Dictionary of numpy arrays for testing."""

    return {
        "mars": np.array([0.5, 0.6, 0.7], dtype=np.float64),
        "venus": np.array([1.0, 1.1], dtype=np.float64),
    }


@pytest.fixture
def numpy_scalar() -> np.float64:
    """Numpy scalar for testing."""

    return np.float64(42.5)


@pytest.fixture
def python_list() -> list[float]:
    """Python list for testing."""
    return [1.0, 2.0, 3.0]


@pytest.fixture
def skycoord_single() -> SkyCoord:
    """Single SkyCoord for testing."""
    from astropy import units as u
    from astropy.coordinates import SkyCoord

    return SkyCoord(ra=10 * u.deg, dec=20 * u.deg)


@pytest.fixture
def skycoord_multiple() -> SkyCoord:
    """Multiple SkyCoords for testing."""
    from astropy import units as u
    from astropy.coordinates import SkyCoord

    return SkyCoord(ra=[10, 20, 30] * u.deg, dec=[20, 30, 40] * u.deg)


@pytest.fixture
def skycoord_comprehensive() -> SkyCoord:
    """SkyCoord for comprehensive model testing."""
    from astropy import units as u
    from astropy.coordinates import SkyCoord

    return SkyCoord(ra=[10, 20] * u.deg, dec=[20, 30] * u.deg)


@pytest.fixture
def visibility_computed_values(altaz_frame: AltAz, angle_array: Any) -> VisibilityComputedValues:
    """VisibilityComputedValues fixture for integration testing."""
    import numpy as np
    from astropy import units as u
    from astropy.coordinates import SkyCoord

    from across.tools.core.schemas.visibility import VisibilityComputedValues

    airmass_values = np.array([1.0, 1.2, 1.5])
    return VisibilityComputedValues(
        sun_angle=angle_array,
        moon_angle=[60.0] * u.deg,  # Use array for consistency
        alt_az=SkyCoord(alt=[30, 45] * u.deg, az=[45, 90] * u.deg, frame=altaz_frame),
        air_mass=airmass_values,
        body_magnitude={"mars": np.array([0.5, 0.6])},
    )


@pytest.fixture
def visibility_computed_values_json(visibility_computed_values: VisibilityComputedValues) -> dict[str, Any]:
    """Serialized VisibilityComputedValues as dict."""
    import json

    json_str = visibility_computed_values.model_dump_json()
    return cast(dict[str, Any], json.loads(json_str))



from typing import Annotated

from pydantic import Field

from .airmass import AirmassConstraint
from .alt_az import AltAzConstraint
from .base import get_slice
from .bright_star import BrightStarConstraint
from .daytime import DaytimeConstraint
from .earth_limb import EarthLimbConstraint
from .ecliptic_latitude import EclipticLatitudeConstraint
from .galactic_bulge import GalacticBulgeConstraint
from .galactic_plane import GalacticPlaneConstraint
from .logical import AndConstraint, NotConstraint, OrConstraint, XorConstraint
from .moon_angle import MoonAngleConstraint
from .pointing import PointingConstraint
from .saa import SAAPolygonConstraint
from .solar_system import SolarSystemConstraint
from .sun_angle import SunAngleConstraint

__all__ = [
    "AllConstraint",
    "get_slice",
    "EarthLimbConstraint",
    "MoonAngleConstraint",
    "SunAngleConstraint",
    "SAAPolygonConstraint",
    "AltAzConstraint",
    "GalacticPlaneConstraint",
    "BrightStarConstraint",
    "AirmassConstraint",
    "EclipticLatitudeConstraint",
    "GalacticBulgeConstraint",
    "SolarSystemConstraint",
    "DaytimeConstraint",
    "AndConstraint",
    "OrConstraint",
    "NotConstraint",
    "XorConstraint",
    "PointingConstraint",
    "Constraint",
]

# Define a type that covers all constraints
AllConstraint = Annotated[
    EarthLimbConstraint
    | MoonAngleConstraint
    | SunAngleConstraint
    | SAAPolygonConstraint
    | AltAzConstraint
    | GalacticPlaneConstraint
    | BrightStarConstraint
    | AirmassConstraint
    | EclipticLatitudeConstraint
    | GalacticBulgeConstraint
    | SolarSystemConstraint
    | DaytimeConstraint
    | AndConstraint
    | OrConstraint
    | NotConstraint
    | XorConstraint
    | PointingConstraint,
    Field(discriminator="name"),
]

Constraint = AllConstraint
# Rebuild models to resolve forward references
AndConstraint.model_rebuild()
OrConstraint.model_rebuild()
NotConstraint.model_rebuild()
XorConstraint.model_rebuild()



import numpy as np
import pytest

from across.tools.core.math import find_duplicates, x_rot, y_rot, z_rot


class TestXRotation:
    """Test suite for x_rot function"""

    @pytest.mark.parametrize(
        "angle,expected",
        [
            (0, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (90, np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])),
            (180, np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])),
            (360, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (-90, np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]])),
        ],
    )
    def test_x_rot_angles(self, angle: float, expected: np.typing.NDArray[np.float64]) -> None:
        """Test rotation matrix for various angles"""
        actual = x_rot(angle)
        np.testing.assert_array_almost_equal(actual, expected)

    def test_x_rot_return_type(self) -> None:
        """Test return type is numpy array"""
        result = x_rot(45)
        assert isinstance(result, np.ndarray)

    def test_x_rot_dtype(self) -> None:
        """Test array dtype is float64"""
        result = x_rot(45)
        assert result.dtype == np.float64


class TestZRotation:
    """Test suite for z_rot function"""

    @pytest.mark.parametrize(
        "angle,expected",
        [
            (0, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (90, np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])),
            (180, np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])),
            (360, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (-90, np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])),
        ],
    )
    def test_z_rot_angles(self, angle: float, expected: np.typing.NDArray[np.float64]) -> None:
        """Test rotation matrix for various angles"""
        actual = z_rot(angle)
        np.testing.assert_array_almost_equal(actual, expected)

    def test_z_rot_return_type(self) -> None:
        """Test return type is numpy array"""
        result = z_rot(45)
        assert isinstance(result, np.ndarray)

    def test_z_rot_dtype(self) -> None:
        """Test array dtype is float64"""
        result = z_rot(45)
        assert result.dtype == np.float64


class TestYRotation:
    """Test suite for y_rot function"""

    @pytest.mark.parametrize(
        "angle,expected",
        [
            (0, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (90, np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])),
            (180, np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])),
            (360, np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])),
            (-90, np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]])),
        ],
    )
    def test_y_rot_angles(self, angle: float, expected: np.typing.NDArray[np.float64]) -> None:
        """Test rotation matrix for various angles"""
        actual = y_rot(angle)
        np.testing.assert_array_almost_equal(actual, expected)

    def test_y_rot_return_type(self) -> None:
        """Test return type is numpy array"""
        result = y_rot(45)
        assert isinstance(result, np.ndarray)

    def test_y_rot_dtype(self) -> None:
        """Test array dtype is float64"""
        result = y_rot(45)
        assert result.dtype == np.float64


class TestFindDuplicates:
    """Test suite for find_duplicates function"""

    def test_find_duplicates_empty_list(self) -> None:
        """Test with empty list"""
        assert find_duplicates([]) == []

    def test_find_duplicates_no_duplicates(self) -> None:
        """Test with list containing no duplicates"""
        assert find_duplicates([1, 2, 3, 4]) == []

    def test_find_duplicates_single_duplicate(self) -> None:
        """Test with list containing single duplicate"""
        assert find_duplicates([1, 2, 2, 3]) == [2]

    def test_find_duplicates_multiple_duplicates(self) -> None:
        """Test with list containing multiple duplicates"""
        assert find_duplicates([1, 1, 2, 2, 3]) == [1, 2]

    def test_find_duplicates_multiple_occurrences(self) -> None:
        """Test with list containing elements appearing more than twice"""
        assert find_duplicates([1, 1, 1, 2, 2, 3]) == [1, 2]

    def test_find_duplicates_return_type(self) -> None:
        """Test return type is list"""
        result = find_duplicates([1, 2, 2, 3])
        assert isinstance(result, list)



from datetime import datetime, timedelta
from typing import Any

import astropy.units as u  # type: ignore[import-untyped]
import pytest
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.ephemeris import Ephemeris
from across.tools.visibility.constraints.solar_system import SolarSystemConstraint


class MockEphemerisWithSun(Ephemeris):
    """Mock ephemeris class for solar system magnitude testing."""

    def __init__(self, sun: SkyCoord) -> None:
        self.sun = sun

    def prepare_data(self) -> None:
        """Mock method to prepare data."""
        pass


@pytest.fixture
def sun_coord() -> SkyCoord:
    """Create a SkyCoord instance for sun position testing."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=1 * u.AU)


@pytest.fixture
def mock_ephemeris_with_sun(sun_coord: SkyCoord) -> MockEphemerisWithSun:
    """Fixture for a mock ephemeris with sun coordinates."""
    sun_array = SkyCoord([sun_coord.ra], [sun_coord.dec], distance=[sun_coord.distance])
    return MockEphemerisWithSun(sun_array)


@pytest.fixture
def body_coord_1au() -> SkyCoord:
    """Fixture for a body coordinate at 1 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=1 * u.AU)


@pytest.fixture
def body_coord_1_5au() -> SkyCoord:
    """Fixture for a body coordinate at 1.5 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=1.5 * u.AU)


@pytest.fixture
def body_coord_5au() -> SkyCoord:
    """Fixture for a body coordinate at 5 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=5 * u.AU)


@pytest.fixture
def body_coord_9au() -> SkyCoord:
    """Fixture for a body coordinate at 9 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=9 * u.AU)


@pytest.fixture
def body_coord_19au() -> SkyCoord:
    """Fixture for a body coordinate at 19 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=19 * u.AU)


@pytest.fixture
def body_coord_30au() -> SkyCoord:
    """Fixture for a body coordinate at 30 AU distance."""
    return SkyCoord(ra=0 * u.deg, dec=0 * u.deg, distance=30 * u.AU)


@pytest.fixture
def mock_get_slice(monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock get_slice to return slice(0, 5) for testing."""
    import across.tools.visibility.constraints.solar_system as ss

    monkeypatch.setattr(ss, "get_slice", lambda time, ephem: slice(0, 5))
    monkeypatch.setattr("across.tools.visibility.constraints.base.get_slice", lambda time, ephem: slice(0, 5))


@pytest.fixture
def mock_get_body(monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock get_body to return dummy SkyCoord for testing."""

    def mock_body(body: Any, time: Time, location: Any) -> SkyCoord:
        num = len(time) if hasattr(time, "__len__") else 1
        return SkyCoord(ra=[150] * num * u.deg, dec=[20] * num * u.deg, distance=[1.5] * num * u.AU)

    monkeypatch.setattr("astropy.coordinates.get_body", mock_body)


@pytest.fixture
def multi_time_array() -> Time:
    """Fixture for a multi-step time array used in combined tests."""
    begin = datetime(2025, 2, 12, 0, 0, 0)
    times = [begin + timedelta(minutes=i * 5) for i in range(5)]
    return Time(times, scale="utc")


@pytest.fixture
def test_coord() -> SkyCoord:
    """Fixture for test coordinate used in combined tests."""
    return SkyCoord(ra=150 * u.deg, dec=20 * u.deg)


@pytest.fixture
def test_constraint() -> SolarSystemConstraint:
    """Fixture for test constraint used in combined tests."""
    return SolarSystemConstraint(bodies=["mars", "jupiter"], min_separation=10.0)


@pytest.fixture
def solar_system_constraint() -> SolarSystemConstraint:
    """Fixture for a basic SolarSystemConstraint instance."""
    return SolarSystemConstraint()


@pytest.fixture
def solar_system_constraint_with_separation() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with custom separation."""
    return SolarSystemConstraint(min_separation=10.0)


@pytest.fixture
def solar_system_constraint_custom() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with custom bodies and separation."""
    return SolarSystemConstraint(min_separation=20.0, bodies=["mars", "jupiter"])


@pytest.fixture
def solar_system_constraint_small_separation() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with small separation."""
    return SolarSystemConstraint(min_separation=1.0)


@pytest.fixture
def solar_system_constraint_large_separation() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with large separation."""
    return SolarSystemConstraint(min_separation=100.0)


@pytest.fixture
def solar_system_constraint_single_body() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with a single body."""
    return SolarSystemConstraint(bodies=["mars"], min_separation=10.0)


@pytest.fixture
def solar_system_constraint_multiple_bodies() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with multiple bodies."""
    return SolarSystemConstraint(bodies=["venus", "mars", "jupiter"], min_separation=10.0)


@pytest.fixture
def solar_system_constraint_empty_bodies() -> SolarSystemConstraint:
    """Fixture for SolarSystemConstraint with no bodies."""
    return SolarSystemConstraint(bodies=[])


@pytest.fixture
def slice_index() -> slice:
    """Fixture for a slice index used in testing."""
    return slice(0, 1)



from enum import Enum


class EnergyUnit(str, Enum):
    """
    Enum to represent the bandpass energy types
    """

    eV = "eV"  # noqa: N815
    keV = "keV"  # noqa: N815
    MeV = "MeV"
    GeV = "GeV"
    TeV = "TeV"



from datetime import timedelta

import astropy.units as u  # type: ignore[import-untyped]
import numpy as np
import pytest
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]
from numpy.typing import NDArray

from across.tools.ephemeris import Ephemeris


class TestEphemeris:
    """Test the Ephemeris class."""

    def test_abc_class_not_implementation(
        self,
        not_implemented_ephemeris_class: type[Ephemeris],
        ephemeris_begin: Time,
        ephemeris_end: Time,
        ephemeris_step_size: TimeDelta,
    ) -> None:
        """Test that the Ephemeris class is an abstract base class."""

        test = not_implemented_ephemeris_class(
            begin=ephemeris_begin, end=ephemeris_end, step_size=ephemeris_step_size
        )
        with pytest.raises(NotImplementedError):
            test.prepare_data()

    def test_compute_ground_ephemeris_instance(self, keck_ground_ephemeris: Ephemeris) -> None:
        """Test that compute_ground_ephemeris returns an Ephemeris object."""
        assert isinstance(keck_ground_ephemeris, Ephemeris)

    def test_compute_ground_ephemeris_timestamp_length(self, keck_ground_ephemeris: Ephemeris) -> None:
        """Test that the timestamp has a length of 6."""
        assert len(keck_ground_ephemeris.timestamp) == 6

    def test_compute_ground_ephemeris_latitude(
        self, keck_ground_ephemeris: Ephemeris, keck_latitude: u.Quantity
    ) -> None:
        """Test the latitude of the ground ephemeris."""
        assert np.isclose(
            keck_ground_ephemeris.earth_location.lat.to_value(u.deg), keck_latitude.to_value(u.deg), 1e-3
        )

    def test_compute_ground_ephemeris_longitude(
        self, keck_ground_ephemeris: Ephemeris, keck_longitude: u.Quantity
    ) -> None:
        """Test the longitude of the ground ephemeris."""
        assert np.isclose(
            keck_ground_ephemeris.earth_location.lon.to_value(u.deg), keck_longitude.to_value(u.deg), 1e-3
        )

    def test_compute_ground_ephemeris_height(
        self, keck_ground_ephemeris: Ephemeris, keck_height: u.Quantity
    ) -> None:
        """Test the height of the ground ephemeris."""
        assert np.isclose(
            keck_ground_ephemeris.earth_location.height.to_value(u.m), keck_height.to_value(u.m), 1e-3
        )

    def test_compute_tle_ephemeris_instance(self, hubble_tle_ephemeris: Ephemeris) -> None:
        """Test that compute_tle_ephemeris returns an Ephemeris object."""
        assert isinstance(hubble_tle_ephemeris, Ephemeris)

    def test_compute_tle_ephemeris_timestamp_length(self, hubble_tle_ephemeris: Ephemeris) -> None:
        """Test that the timestamp has a length of 6."""
        assert len(hubble_tle_ephemeris.timestamp) == 6

    def test_compute_tle_ephemeris_gcrs(
        self, hubble_tle_ephemeris: Ephemeris, hubble_gcrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the GCRS coordinates from the TLE ephemeris."""
        assert np.allclose(hubble_tle_ephemeris.gcrs.cartesian.xyz.to_value(u.km), hubble_gcrs_value_km, 1e-3)

    def test_compute_tle_ephemeris_itrs(
        self, hubble_tle_ephemeris: Ephemeris, hubble_itrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the ITRS coordinates from the TLE ephemeris."""
        assert np.allclose(
            hubble_tle_ephemeris.earth_location.itrs.cartesian.xyz.to_value(u.km), hubble_itrs_value_km, 1e-3
        )

    def test_compute_jpl_ephemeris_instance(self, hubble_jpl_ephemeris: Ephemeris) -> None:
        """Test that compute_jpl_ephemeris returns an Ephemeris object."""
        assert isinstance(hubble_jpl_ephemeris, Ephemeris)

    def test_compute_jpl_ephemeris_timestamp_length(self, hubble_jpl_ephemeris: Ephemeris) -> None:
        """Test that the timestamp has a length of 6."""
        assert len(hubble_jpl_ephemeris.timestamp) == 6

    def test_compute_jpl_ephemeris_gcrs(
        self, hubble_jpl_ephemeris: Ephemeris, hubble_gcrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the GCRS coordinates from the JPL ephemeris."""
        assert np.allclose(hubble_jpl_ephemeris.gcrs.cartesian.xyz.to_value(u.km), hubble_gcrs_value_km, 1e-3)

    def test_compute_jpl_ephemeris_itrs(
        self, hubble_jpl_ephemeris: Ephemeris, hubble_itrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the ITRS coordinates from the JPL ephemeris."""
        assert np.allclose(
            hubble_jpl_ephemeris.earth_location.itrs.cartesian.xyz.to_value(u.km), hubble_itrs_value_km, 1
        )

    def test_compute_spice_ephemeris_instance(self, hubble_spice_ephemeris: Ephemeris) -> None:
        """Test that compute_spice_ephemeris returns an Ephemeris object."""
        assert isinstance(hubble_spice_ephemeris, Ephemeris)

    def test_compute_spice_ephemeris_timestamp_length(self, hubble_spice_ephemeris: Ephemeris) -> None:
        """Test that the timestamp has a length of 6."""
        assert len(hubble_spice_ephemeris.timestamp) == 6

    def test_compute_spice_ephemeris_gcrs(
        self, hubble_spice_ephemeris: Ephemeris, hubble_gcrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the GCRS coordinates from the SPICE ephemeris."""
        assert np.allclose(
            hubble_spice_ephemeris.gcrs.cartesian.xyz.to_value(u.km), hubble_gcrs_value_km, 1e-3
        )

    def test_compute_spice_ephemeris_itrs(
        self, hubble_spice_ephemeris: Ephemeris, hubble_itrs_value_km: NDArray[np.float64]
    ) -> None:
        """Test the ITRS coordinates from the SPICE ephemeris."""
        assert np.allclose(
            hubble_spice_ephemeris.earth_location.itrs.cartesian.xyz.to_value(u.km), hubble_itrs_value_km, 1
        )

    def test_ephemeris_init_time_objects_begin(
        self,
        mock_ephemeris: Ephemeris,
        ephemeris_begin: Time,
    ) -> None:
        """Test the initialization of the Ephemeris class with Time objects - begin."""
        assert mock_ephemeris.begin == ephemeris_begin

    def test_ephemeris_init_time_objects_end(
        self,
        mock_ephemeris: Ephemeris,
        ephemeris_end: Time,
    ) -> None:
        """Test the initialization of the Ephemeris class with Time objects - end."""
        assert mock_ephemeris.end == ephemeris_end

    def test_ephemeris_init_time_objects_step_size(
        self,
        mock_ephemeris: Ephemeris,
        ephemeris_step_size: TimeDelta,
    ) -> None:
        """Test the initialization of the Ephemeris class with Time objects - step size."""
        assert mock_ephemeris.step_size == ephemeris_step_size

    def test_ephemeris_init_datetime_objects_begin(self, mock_ephemeris: Ephemeris) -> None:
        """Test the initialization of the Ephemeris class with datetime objects - begin."""
        assert isinstance(mock_ephemeris.begin, Time)

    def test_ephemeris_init_datetime_objects_end(self, mock_ephemeris: Ephemeris) -> None:
        """Test the initialization of the Ephemeris class with datetime objects - end."""
        assert isinstance(mock_ephemeris.end, Time)

    def test_ephemeris_init_datetime_objects_step_size(self, mock_ephemeris: Ephemeris) -> None:
        """Test the initialization of the Ephemeris class with datetime objects - step size."""
        assert mock_ephemeris.step_size == TimeDelta(60 * u.s)

    def test_ephemeris_init_step_size_int(self, mock_ephemeris: Ephemeris) -> None:
        """Test the initialization of the Ephemeris class with int step size."""
        assert mock_ephemeris.step_size == TimeDelta(60 * u.s)

    def test_ephemeris_len(self, mock_ephemeris: Ephemeris) -> None:
        """Test the length of the Ephemeris class."""
        assert len(mock_ephemeris) == 6

    def test_ephemeris_len_single_date(
        self, mock_ephemeris_class: type[Ephemeris], ephemeris_begin: Time, ephemeris_step_size: TimeDelta
    ) -> None:
        """Test the length of the Ephemeris class when begin and end are the same."""
        ephem = mock_ephemeris_class(ephemeris_begin, ephemeris_begin, ephemeris_step_size)
        assert len(ephem) == 1

    def test_ephemeris_index(
        self, ephemeris_begin: Time, mock_ephemeris: Ephemeris, ephemeris_step_size: TimeDelta
    ) -> None:
        """Test the index method of the Ephemeris class."""
        t = ephemeris_begin + ephemeris_step_size * 3
        assert mock_ephemeris.index(t) == 3

    def test_ephemeris_index_outside_range(self, mock_ephemeris: Ephemeris) -> None:
        """Test the index method of the Ephemeris class with a time outside the
        range."""
        t = Time("2023-01-01T00:05:00")
        with pytest.raises(AssertionError):
            mock_ephemeris.index(t)

    def test_compute_timestamp_length(
        self, ephemeris_begin: Time, ephemeris_end: Time, mock_ephemeris: Ephemeris
    ) -> None:
        """Test the length of timestamps from _compute_timestamp method."""
        timestamps = mock_ephemeris._compute_timestamp()
        assert len(timestamps) == 6

    def test_compute_timestamp_first_value(
        self, ephemeris_begin: Time, ephemeris_end: Time, mock_ephemeris: Ephemeris
    ) -> None:
        """Test the first timestamp value from _compute_timestamp method."""
        timestamps = mock_ephemeris._compute_timestamp()
        assert timestamps[0] == ephemeris_begin

    def test_compute_timestamp_last_value(
        self, ephemeris_begin: Time, ephemeris_end: Time, mock_ephemeris: Ephemeris
    ) -> None:
        """Test the last timestamp value from _compute_timestamp method."""
        timestamps = mock_ephemeris._compute_timestamp()
        assert timestamps[-1] == ephemeris_end

    def test_compute_timestamp_single_length(
        self, mock_ephemeris_class: type[Ephemeris], ephemeris_begin: Time, ephemeris_step_size: TimeDelta
    ) -> None:
        """Test the length of timestamps when begin equals end."""
        ephem = mock_ephemeris_class(ephemeris_begin, ephemeris_begin, ephemeris_step_size)
        timestamps = ephem._compute_timestamp()
        assert len(timestamps) == 1

    def test_compute_timestamp_single_value(
        self, mock_ephemeris_class: type[Ephemeris], ephemeris_begin: Time, ephemeris_step_size: TimeDelta
    ) -> None:
        """Test the timestamp value when begin equals end."""
        ephem = mock_ephemeris_class(ephemeris_begin, ephemeris_begin, ephemeris_step_size)
        timestamps = ephem._compute_timestamp()
        assert timestamps[0] == ephemeris_begin

    def test_prepare_data(self, mock_ephemeris: Ephemeris) -> None:
        """Test the prepare_data method."""

        mock_ephemeris.prepare_data()  # Call prepare_data to ensure it runs without errors
        assert True  # If it runs without errors, the test passes

    def test_step_size_parameter_is_int(
        self, mock_ephemeris_class: type[Ephemeris], ephemeris_begin: Time, ephemeris_end: Time
    ) -> None:
        """Test the step_size attribute with an int."""
        ephem = mock_ephemeris_class(begin=ephemeris_begin, end=ephemeris_end, step_size=60)
        assert ephem.step_size == TimeDelta(60 * u.s)

    def test_step_size_parameter_is_datetime(
        self, mock_ephemeris_class: type[Ephemeris], ephemeris_begin: Time, ephemeris_end: Time
    ) -> None:
        """Test the step_size attribute with a datetime."""
        ephem = mock_ephemeris_class(
            begin=ephemeris_begin, end=ephemeris_end, step_size=timedelta(seconds=60)
        )
        assert ephem.step_size == TimeDelta(60 * u.s)

    @pytest.mark.parametrize("i", range(6))
    def test_every_index_near_values(self, mock_ephemeris: Ephemeris, i: int) -> None:
        """Test the index method for every timestamp with a small random offset."""
        t = mock_ephemeris.timestamp[i] + TimeDelta((0.2 ** np.random.random() - 0.1) * u.s)
        assert mock_ephemeris.index(t) == i



from datetime import datetime, timedelta

import astropy.units as u  # type: ignore[import-untyped]
import pytest
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]

from across.tools.core.schemas import Coordinate, Polygon
from across.tools.footprint import Footprint, Pointing
from across.tools.visibility.constraints.pointing import PointingConstraint


@pytest.fixture
def origin_sky_coord() -> SkyCoord:
    """Create a SkyCoord located at the center of the test pointing footprint."""
    return SkyCoord(ra=10 * u.deg, dec=10 * u.deg)


@pytest.fixture
def pointing_constraint(ephemeris_begin: datetime, ephemeris_end: datetime) -> PointingConstraint:
    """Create a PointingConstraint with a simple square footprint around origin_sky_coord."""
    footprint = Footprint(
        detectors=[
            Polygon(
                coordinates=[
                    Coordinate(ra=9.0, dec=9.0),
                    Coordinate(ra=11.0, dec=9.0),
                    Coordinate(ra=11.0, dec=11.0),
                    Coordinate(ra=9.0, dec=11.0),
                ]
            )
        ]
    )

    pointing = Pointing(
        footprint=footprint,
        start_time=ephemeris_begin - timedelta(minutes=1),
        end_time=ephemeris_end + timedelta(minutes=1),
    )

    return PointingConstraint(pointings=[pointing])



from datetime import datetime
from unittest.mock import patch

import astropy.units as u  # type: ignore[import-untyped]  # type: ignore[import-untyped]
import numpy as np
import pytest
from astropy.coordinates import (  # type: ignore[import-untyped]
    Latitude,
    Longitude,
)
from astropy.table import Table  # type: ignore[import-untyped]
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]
from astropy.units import Quantity as uQuantity
from numpy.typing import NDArray

from across.tools.core.schemas.tle import TLE
from across.tools.ephemeris import (
    Ephemeris,
    compute_ground_ephemeris,
    compute_jpl_ephemeris,
    compute_spice_ephemeris,
    compute_tle_ephemeris,
)
from across.tools.ephemeris.ground_ephem import GroundEphemeris


class MockEphemeris(Ephemeris):
    """Mock class for testing the Ephemeris class."""

    def prepare_data(self) -> None:
        """Mock method to prepare data."""
        pass


class NotImplementedEphemeris(Ephemeris):
    """Mock class for testing the Ephemeris class."""

    def prepare_data(self) -> None:
        """Mock method to prepare data."""
        super().prepare_data()  # type: ignore[safe-super]


@pytest.fixture
def mock_ephemeris_class() -> type[MockEphemeris]:
    """Fixture for the MockEphemeris class."""
    return MockEphemeris


@pytest.fixture
def not_implemented_ephemeris_class() -> type[NotImplementedEphemeris]:
    """Fixture for the NotImplementedEphemeris class."""
    return NotImplementedEphemeris


@pytest.fixture
def ephemeris_begin() -> Time:
    """Fixture to provide a begin datetime for testing."""
    return Time("2025-02-12 00:00:00", scale="utc")


@pytest.fixture
def ephemeris_end() -> Time:
    """Fixture to provide an end datetime for testing."""
    return Time("2025-02-12 00:05:00", scale="utc")


@pytest.fixture
def mock_ephemeris(
    ephemeris_begin: Time, ephemeris_end: Time, ephemeris_step_size: TimeDelta
) -> MockEphemeris:
    """Fixture to provide a MockEphemeris instance."""
    return MockEphemeris(begin=ephemeris_begin, end=ephemeris_end, step_size=ephemeris_step_size)


@pytest.fixture
def ephemeris_step_size() -> TimeDelta:
    """Fixture to provide a step_size for testing."""
    return TimeDelta(60 * u.s)


@pytest.fixture
def keck_latitude() -> u.Quantity:
    """Fixture to provide a latitude for testing."""
    return 20.879 * u.deg


@pytest.fixture
def keck_longitude() -> u.Quantity:
    """Fixture to provide a longitude for testing."""
    return 155.6655 * u.deg


@pytest.fixture
def keck_height() -> u.Quantity:
    """Fixture to provide a height for testing."""
    return 4160 * u.m


@pytest.fixture
def hubble_tle() -> TLE:
    """Fixture to provide a TLE for the Hubble Space Telescope."""
    return TLE(
        satellite_name="HST",
        tle1="1 20580U 90037B   25043.65115181  .00008438  00000+0  34020-3 0  9990",
        tle2="2 20580  28.4676 167.7952 0002864 136.5792 223.5028 15.22991033713929",
    )


@pytest.fixture
def hubble_naif_id() -> int:
    """Fixture to provide the NAIF ID for the Hubble Space Telescope."""
    return -48


@pytest.fixture
def hubble_gcrs_value_km() -> NDArray[np.float64]:
    """Fixture to provide the GCRS position of the Hubble satellite."""
    return np.array(
        [
            [-6601.79957494, -6463.92010459, -6297.43794083, -6103.08764513, -5881.72745354, -5634.33546338],
            [-1415.26655239, -1803.68208289, -2184.11634725, -2554.88519298, -2914.34708129, -3260.9103933],
            [1281.48161724, 1479.25441952, 1670.46341691, 1854.25957144, 2029.826681, 2196.38502342],
        ]
    )


@pytest.fixture
def hubble_itrs_value_km() -> NDArray[np.float64]:
    """Fixture to provide the Earth location of the Hubble satellite."""
    return np.array(
        [
            [4331.13875516, 4007.24602751, 3667.39060024, 3312.88335429, 2945.09455093, 2565.44849455],
            [5183.55079613, 5387.41092415, 5570.43884626, 5731.96023048, 5871.38387356, 5988.20382911],
            [1265.29459209, 1463.38942926, 1654.99066752, 1839.24753875, 2015.3418068, 2182.49142062],
        ]
    )


@pytest.fixture
def hubble_horizons_vectors() -> Table:
    """Fixture to provide the Hubble Space Telescope vectors from
    astroquery.jplephemeris's Horizons.vectors method."""

    return Table(
        [
            np.array(
                [
                    "Hubble Space Telescope (spacecraft)",
                    "Hubble Space Telescope (spacecraft)",
                    "Hubble Space Telescope (spacecraft)",
                    "Hubble Space Telescope (spacecraft)",
                    "Hubble Space Telescope (spacecraft)",
                    "Hubble Space Telescope (spacecraft)",
                ]
            ),
            np.array(
                [
                    2460718.500800752,
                    2460718.501495197,
                    2460718.502189641,
                    2460718.502884086,
                    2460718.50357853,
                    2460718.504272975,
                ]
            )
            * u.d,
            np.array(
                [
                    "A.D. 2025-Feb-12 00:01:09.1850",
                    "A.D. 2025-Feb-12 00:02:09.1850",
                    "A.D. 2025-Feb-12 00:03:09.1850",
                    "A.D. 2025-Feb-12 00:04:09.1850",
                    "A.D. 2025-Feb-12 00:05:09.1850",
                    "A.D. 2025-Feb-12 00:06:09.1850",
                ]
            ),
            np.array(
                np.array(
                    [
                        -4.413109685504098e-05,
                        -4.320951686348332e-05,
                        -4.209674074411141e-05,
                        -4.079767809571204e-05,
                        -3.931806519583499e-05,
                        -3.76644395176696e-05,
                    ]
                )
            )
            * u.AU,
            np.array(
                [
                    -9.459679304384228e-06,
                    -1.20560763138425e-05,
                    -1.459912674999905e-05,
                    -1.707757297872861e-05,
                    -1.948044216827302e-05,
                    -2.179709512380254e-05,
                ]
            )
            * u.AU,
            np.array(
                [
                    8.565141089418774e-06,
                    9.887171518916822e-06,
                    1.116533134540432e-05,
                    1.239394518573508e-05,
                    1.356755710675922e-05,
                    1.468095498206593e-05,
                ]
            )
            * u.AU,
            np.array(
                [
                    0.001187350603237076,
                    0.001465801721275105,
                    0.001737789196464366,
                    0.002002106051621462,
                    0.002257579357790702,
                    0.002503075493910799,
                ]
            )
            * u.AU
            / u.d,
            np.array(
                [
                    -0.003771715115954226,
                    -0.003703120624754096,
                    -0.003618134112267888,
                    -0.003517129872519882,
                    -0.003400553506769829,
                    -0.003268919895516453,
                ]
            )
            * u.AU
            / u.d,
            np.array(
                [
                    0.001932503216081898,
                    0.001873512471782746,
                    0.001806203302310515,
                    0.001730873692075622,
                    0.001647857439270663,
                    0.001557522639078362,
                ]
            )
            * u.AU
            / u.d,
            np.array(
                [
                    2.653221126234431e-07,
                    2.653074405706743e-07,
                    2.652928068499324e-07,
                    2.652782970600778e-07,
                    2.652639978997604e-07,
                    2.652499963858138e-07,
                ]
            )
            * u.d,
            np.array(
                [
                    4.593909973053949e-05,
                    4.593655934335226e-05,
                    4.593402559314974e-05,
                    4.593151330091518e-05,
                    4.592903747805448e-05,
                    4.592661319106531e-05,
                ]
            )
            * u.AU,
            np.array(
                [
                    -3.650623268807726e-06,
                    -3.651561928277356e-06,
                    -3.631391291522598e-06,
                    -3.589746103254491e-06,
                    -3.526451748811547e-06,
                    -3.441531501265534e-06,
                ]
            )
            * u.AU
            / u.d,
        ],
        names=(
            "targetname",
            "datetime_jd",
            "datetime_str",
            "x",
            "y",
            "z",
            "vx",
            "vy",
            "vz",
            "lighttime",
            "range",
            "range_rate",
        ),
    )


@pytest.fixture
def hubble_mocked_spkezr() -> NDArray[np.float64]:
    """Fixture to provide the SPK ephemeris states for the Hubble Telescope."""

    return np.array(
        [
            [-6.60191801e03, -1.41514824e03, 1.28132705e03, 2.05584676e00, -6.53056183e00, 3.34604582e00],
            [-6.46405158e03, -1.80356370e03, 1.47909998e03, 2.53797285e00, -6.41179340e00, 3.24390588e00],
            [-6.29758261e03, -2.18399862e03, 1.67030997e03, 3.00890741e00, -6.26464290e00, 3.12736294e00],
            [-6.10324558e03, -2.55476889e03, 1.85410797e03, 3.46656062e00, -6.08975839e00, 2.99693296e00],
            [-5.88189862e03, -2.91423299e03, 2.02967781e03, 3.90890161e00, -5.88791142e00, 2.85319389e00],
            [-5.63451972e03, -3.26079933e03, 2.19623975e03, 4.33396756e00, -5.65999348e00, 2.69678307e00],
        ]
    )


@pytest.fixture
def hubble_spice_kernel_url() -> str:
    """Fixture to provide the spice kernel URL for the Hubble Space Telescope."""
    return "https://naif.jpl.nasa.gov/pub/naif/HST/kernels/spk/hst.bsp"


@pytest.fixture
def keck_ground_ephemeris(
    ephemeris_begin: datetime,
    ephemeris_end: datetime,
    ephemeris_step_size: int,
    keck_latitude: u.Quantity,
    keck_longitude: u.Quantity,
    keck_height: u.Quantity,
) -> Ephemeris:
    """Fixture to provide the ground ephemeris for the Keck Observatory."""
    return compute_ground_ephemeris(
        begin=ephemeris_begin,
        end=ephemeris_end,
        step_size=ephemeris_step_size,
        latitude=keck_latitude,
        longitude=keck_longitude,
        height=keck_height,
    )


@pytest.fixture
def hubble_tle_ephemeris(
    ephemeris_begin: datetime, ephemeris_end: datetime, ephemeris_step_size: int, hubble_tle: TLE
) -> Ephemeris:
    """Fixture to provide the TLE ephemeris for the Hubble satellite."""
    return compute_tle_ephemeris(
        begin=ephemeris_begin,
        end=ephemeris_end,
        step_size=ephemeris_step_size,
        tle=hubble_tle,
    )


@pytest.fixture
def hubble_spice_ephemeris(
    ephemeris_begin: datetime,
    ephemeris_end: datetime,
    ephemeris_step_size: int,
    hubble_naif_id: int,
    hubble_spice_kernel_url: str,
    hubble_mocked_spkezr: NDArray[np.float64],
) -> Ephemeris:
    """Fixture to provide the spice ephemeris for the Hubble Space Telescope."""
    mock_start_et = 7.92590469e08
    mock_end_et = 7.92590769e08

    with (
        patch("across.tools.ephemeris.spice_ephem.spice.furnsh") as mock_furnsh,
        patch("across.tools.ephemeris.spice_ephem.download_file") as mock_download_file,
        patch("across.tools.ephemeris.spice_ephem.spice.str2et") as mock_str2et,
        patch("across.tools.ephemeris.spice_ephem.spice.spkezr") as mock_spkezr,
        patch("across.tools.ephemeris.spice_ephem.spice.str2et") as mock_str2et,
    ):
        mock_furnsh.return_value = None
        mock_download_file.return_value = None
        mock_str2et.return_value = None
        mock_spkezr.side_effect = [(v,) for v in hubble_mocked_spkezr]
        mock_str2et.side_effect = [mock_start_et, mock_end_et]

        ephemeris = compute_spice_ephemeris(
            begin=ephemeris_begin,
            end=ephemeris_end,
            step_size=ephemeris_step_size,
            naif_id=hubble_naif_id,
            spice_kernel_url=hubble_spice_kernel_url,
        )
        assert mock_furnsh.called
        assert mock_download_file.called
        assert mock_str2et.called
        assert mock_spkezr.called
        return ephemeris


@pytest.fixture
def hubble_jpl_ephemeris(
    ephemeris_begin: datetime,
    ephemeris_end: datetime,
    ephemeris_step_size: int,
    hubble_naif_id: int,
    hubble_horizons_vectors: Table,
) -> Ephemeris:
    """Fixture to provide the JPL ephemeris for the Hubble satellite."""
    with patch("across.tools.ephemeris.jpl_ephem.jpl.Horizons") as mock_horizons_class:
        mock_instance = mock_horizons_class.return_value
        mock_instance.vectors.return_value = hubble_horizons_vectors

        return compute_jpl_ephemeris(
            begin=ephemeris_begin,
            end=ephemeris_end,
            step_size=ephemeris_step_size,
            naif_id=hubble_naif_id,
        )


@pytest.fixture
def ground_ephemeris_params() -> tuple[str, str, int, Latitude, Longitude, uQuantity]:
    """Fixture for ground ephemeris parameters."""
    begin = "2023-01-01T00:00:00"
    end = "2023-01-01T00:01:00"
    step_size = 60
    latitude = Latitude(34.2 * u.deg)
    longitude = Longitude(-118.2 * u.deg)
    height = 100 * u.m
    return begin, end, step_size, latitude, longitude, height


@pytest.fixture
def ground_ephemeris(
    ground_ephemeris_params: tuple[str, str, int, Latitude, Longitude, uQuantity],
) -> GroundEphemeris:
    """Fixture for a GroundEphemeris object with prepared data."""
    begin, end, step_size, latitude, longitude, height = ground_ephemeris_params
    ephemeris = GroundEphemeris(begin, end, step_size, latitude, longitude, height)
    ephemeris.prepare_data()
    return ephemeris



from enum import Enum


class ConstraintType(str, Enum):
    """
    Represents a constraint.
    """

    SUN = "Sun Angle"
    MOON = "Moon Angle"
    EARTH = "Earth Limb"
    WINDOW = "Window"
    UNKNOWN = "Unknown"
    SAA = "South Atlantic Anomaly"
    ALT_AZ = "Altitude/Azimuth Avoidance"
    GALACTIC_PLANE = "Galactic Plane Avoidance"
    BRIGHT_STAR = "Bright Star Avoidance"
    AIRMASS = "Airmass Limit"
    ECLIPTIC_LATITUDE = "Ecliptic Latitude"
    GALACTIC_BULGE = "Galactic Bulge Avoidance"
    SOLAR_SYSTEM = "Solar System Object Avoidance"
    DAYTIME = "Daytime Avoidance"
    TEST = "Test Constraint"
    AND = "And"
    OR = "Or"
    NOT = "Not"
    XOR = "Xor"
    POINTING = "Pointing"



from typing import Literal

import numpy as np
import numpy.typing as npt
from astropy.coordinates import AltAz, SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]
from pydantic import Field

from ...core.enums import ConstraintType, TwilightType
from ...ephemeris import Ephemeris, GroundEphemeris
from .base import ConstraintABC, get_slice


class DaytimeConstraint(ConstraintABC):
    """
    Constraint that avoids daytime observations based on twilight definitions.

    For ground-based telescopes, daytime is defined by the Sun's position relative
    to the horizon. Different twilight types provide different levels of darkness.

    For space-based telescopes, daytime refers to periods when the spacecraft
    is in direct sunlight (requires appropriate ephemeris data).

    The telescope type (ground vs space) is automatically detected based on the
    ephemeris type: GroundEphemeris for ground-based, all others for space-based.

    Parameters
    ----------
    twilight_type : TwilightType
        The type of twilight to use for defining daytime boundaries.
        - 'astronomical': Sun 12-18° below horizon (default)
        - 'nautical': Sun 6-12° below horizon
        - 'civil': Sun 0-6° below horizon
        - 'day': Sun above horizon

    Methods
    -------
    __call__(time, ephemeris, coordinate)
        Checks if the observation time qualifies as daytime for the given constraints.
    """

    name: Literal[ConstraintType.DAYTIME] = ConstraintType.DAYTIME
    short_name: Literal["Daytime"] = "Daytime"
    twilight_type: TwilightType = Field(
        default=TwilightType.ASTRONOMICAL, description="Type of twilight defining daytime boundaries"
    )

    def __call__(self, time: Time, ephemeris: Ephemeris, coordinate: SkyCoord) -> npt.NDArray[np.bool_]:
        """
        Check if the observation time is during daytime.

        Parameters
        ----------
        time : Time
            The time of observation.
        ephemeris : Ephemeris
            The ephemeris (used for Earth location and potentially sunlight data).
        coordinate : SkyCoord
            The coordinate (not used for daytime calculations).

        Returns
        -------
        npt.NDArray[np.bool_]
            Boolean array where True indicates daytime (constraint violated).
        """
        # Auto-detect telescope type based on ephemeris type
        if isinstance(ephemeris, GroundEphemeris):
            return self._check_ground_daytime(time, ephemeris)
        else:
            # Assume space-based for all other ephemeris types
            return self._check_space_daytime(time, ephemeris)

    def _check_ground_daytime(self, time: Time, ephemeris: Ephemeris) -> npt.NDArray[np.bool_]:
        """
        Check daytime for ground-based telescopes using Sun altitude.
        """
        # Find the slice of ephemeris data we need
        i = get_slice(time, ephemeris)

        # Get Sun position
        sun_altaz = ephemeris.sun[i].transform_to(AltAz(obstime=time[i], location=ephemeris.earth_location))

        self.computed_values.sun_altitude = sun_altaz.alt

        # Define daytime based on twilight type
        if self.twilight_type == TwilightType.SUNSET:
            # Twilight starts after sunset
            in_constraint = self.computed_values.sun_altitude.deg > -0.866
        elif self.twilight_type == TwilightType.CIVIL:
            # Sun within 6° of horizon (civil twilight)
            in_constraint = self.computed_values.sun_altitude.deg > -6
        elif self.twilight_type == TwilightType.NAUTICAL:
            # Sun within 12° of horizon (nautical twilight)
            in_constraint = self.computed_values.sun_altitude.deg > -12
        elif self.twilight_type == TwilightType.ASTRONOMICAL:
            # Sun within 18° of horizon (astronomical twilight)
            in_constraint = self.computed_values.sun_altitude.deg > -18

        return np.asarray(in_constraint, dtype=bool)

    def _check_space_daytime(self, time: Time, ephemeris: Ephemeris) -> npt.NDArray[np.bool_]:
        """
        Check daytime for space-based telescopes using proper eclipse calculations.

        Determines if the spacecraft is in direct sunlight by calculating whether
        the Earth is blocking the line of sight to the Sun, accounting for the
        finite sizes of both Earth and Sun.
        """
        # Find the slice of ephemeris data we need
        i = get_slice(time, ephemeris)

        # Calculate angular separation between Sun and Earth centers as seen from spacecraft
        sun_earth_separation = ephemeris.sun[i].separation(ephemeris.earth[i])

        # Spacecraft is in eclipse (not in sunlight) if the angular separation
        # between Sun and Earth centers is less than the sum of their angular radii
        # This means the Earth is blocking the Sun from the spacecraft's perspective
        in_eclipse = sun_earth_separation < ephemeris.earth_radius_angle[i] + ephemeris.sun_radius_angle[i]

        # For space telescopes, daytime means the spacecraft is in sunlight
        # So in_constraint is True when NOT in eclipse (i.e., in sunlight)
        in_constraint = ~in_eclipse

        return np.asarray(in_constraint, dtype=bool)



import pytest
from astropy.coordinates import SkyCoord  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from across.tools.ephemeris import Ephemeris
from across.tools.visibility.constraints.galactic_bulge import GalacticBulgeConstraint


class TestGalacticBulgeConstraintAttributes:
    """Test suite for GalacticBulgeConstraint attributes."""

    def test_constraint_short_name(self) -> None:
        """Test constraint short_name attribute."""
        constraint = GalacticBulgeConstraint()
        assert constraint.short_name == "Galactic Bulge"

    def test_constraint_name_value(self) -> None:
        """Test constraint name.value attribute."""
        constraint = GalacticBulgeConstraint()
        assert constraint.name.value == "Galactic Bulge Avoidance"


class TestGalacticBulgeConstraintInitialization:
    """Test suite for GalacticBulgeConstraint initialization."""

    def test_constraint_initialization_default_min_separation(self) -> None:
        """Test constraint initialization with default min_separation."""
        constraint = GalacticBulgeConstraint()
        assert constraint.min_separation == 10.0

    def test_constraint_initialization_custom_min_separation(self) -> None:
        """Test constraint initialization with custom min_separation."""
        constraint = GalacticBulgeConstraint(min_separation=20.0)
        assert constraint.min_separation == 20.0

    def test_constraint_initialization_zero_min_separation_raises_error(self) -> None:
        """Test constraint initialization with zero min_separation raises error."""
        with pytest.raises(ValueError):
            GalacticBulgeConstraint(min_separation=0.0)  # Must be > 0

    def test_constraint_initialization_negative_min_separation_raises_error(self) -> None:
        """Test constraint initialization with negative min_separation raises error."""
        with pytest.raises(ValueError):
            GalacticBulgeConstraint(min_separation=-5.0)  # Must be > 0


class TestGalacticBulgeConstraintCall:
    """Test suite for GalacticBulgeConstraint __call__ method."""

    def test_constraint_returns_array_like(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        sky_coord: SkyCoord,
        galactic_bulge_constraint: GalacticBulgeConstraint,
    ) -> None:
        """Test that constraint returns array-like result."""
        result = galactic_bulge_constraint(begin_time_array, ground_ephemeris, sky_coord)
        # Result should be array-like (even if single element)
        assert hasattr(result, "dtype")

    def test_constraint_returns_bool_dtype(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        sky_coord: SkyCoord,
        galactic_bulge_constraint: GalacticBulgeConstraint,
    ) -> None:
        """Test that constraint returns boolean dtype."""
        result = galactic_bulge_constraint(begin_time_array, ground_ephemeris, sky_coord)
        assert result.dtype == bool

    def test_constraint_no_violation_when_far_from_bulge(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        far_from_bulge_coord: SkyCoord,
        galactic_bulge_constraint: GalacticBulgeConstraint,
    ) -> None:
        """Test constraint returns False when coordinate is far from galactic bulge."""
        result = galactic_bulge_constraint(begin_time_array, ground_ephemeris, far_from_bulge_coord)
        # Should not be constrained (far from bulge)
        assert not result

    def test_constraint_violation_when_close_to_bulge(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        near_bulge_coord: SkyCoord,
        galactic_bulge_constraint: GalacticBulgeConstraint,
    ) -> None:
        """Test constraint returns True when coordinate is close to galactic bulge."""
        result = galactic_bulge_constraint(begin_time_array, ground_ephemeris, near_bulge_coord)
        # Should be constrained (within 10° of bulge)
        assert result

    def test_constraint_at_bulge_center(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        bulge_center_coord: SkyCoord,
        galactic_bulge_constraint: GalacticBulgeConstraint,
    ) -> None:
        """Test constraint with coordinate at galactic bulge center."""
        result = galactic_bulge_constraint(begin_time_array, ground_ephemeris, bulge_center_coord)
        # Should always be constrained (0° separation < 10°)
        assert result

    def test_constraint_with_small_min_separation_at_bulge_center(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        bulge_center_coord: SkyCoord,
    ) -> None:
        """Test constraint with small min_separation at bulge center."""
        constraint_small = GalacticBulgeConstraint(min_separation=1.0)  # 0° < 1°, constrained
        result_small = constraint_small(begin_time_array, ground_ephemeris, bulge_center_coord)

        assert result_small

    def test_constraint_with_large_min_separation_at_bulge_center(
        self,
        begin_time_array: Time,
        ground_ephemeris: Ephemeris,
        bulge_center_coord: SkyCoord,
    ) -> None:
        """Test constraint with large min_separation at bulge center."""
        constraint_large = GalacticBulgeConstraint(min_separation=30.0)  # 0° < 30°, constrained
        result_large = constraint_large(begin_time_array, ground_ephemeris, bulge_center_coord)

        assert result_large



from uuid import UUID

from pydantic import Field

from ..enums.constraint_type import ConstraintType
from .base import BaseSchema
from .custom_types import AstropyAltAz, AstropyAngles, AstropyDateTime, AstropySkyCoords, NumpyArray


class ConstrainedDate(BaseSchema):
    """
    Represents a constrained date with associated constraint information.

    Attributes
    ----------
    datetime
        The time of the constrained date.
    constraint
        The type of constraint active at this time.
    observatory_id
        UUID of the observatory or instrument associated with this constraint.

    """

    datetime: AstropyDateTime
    constraint: ConstraintType
    observatory_id: UUID


class Window(BaseSchema):
    """
    Represents a time window bounded by constrained dates.

    Attributes
    ----------
    begin
        The start of the window with its associated constraint.
    end
        The end of the window with its associated constraint.

    """

    begin: ConstrainedDate
    end: ConstrainedDate


class ConstraintReason(BaseSchema):
    """
    Represents the reasons for constraints at the boundaries of a visibility window.

    Attributes
    ----------
    start_reason
        Human-readable description of the constraint that ends before the window starts.
    end_reason
        Human-readable description of the constraint that begins after the window ends.

    """

    start_reason: str
    end_reason: str


class VisibilityWindow(BaseSchema):
    """
    Represents a visibility window with duration and constraint information.

    Attributes
    ----------
    window
        The time window defining the start and end of the visibility period.
    max_visibility_duration
        Maximum visibility duration in seconds for this window.
    constraint_reason
        The reasons for constraints at the window boundaries.

    """

    window: Window
    max_visibility_duration: int
    constraint_reason: ConstraintReason


class VisibilityComputedValues(BaseSchema):
    """
    A class to hold computed values used in constraint calculations.

    Attributes
    ----------
    sun_angle
        Angular distance between the Sun and the target coordinate.
    moon_angle
        Angular distance between the Moon and the target coordinate.
    earth_angle
        Angular distance between the Earth limb and the target coordinate.
    alt_az
        Altitude-azimuth coordinates of the target from the observatory.

    """

    sun_angle: AstropyAngles | None = Field(
        default=None, description="Angular distance between the Sun and the coordinate"
    )
    moon_angle: AstropyAngles | None = Field(
        default=None, description="Angular distance between the Moon and the coordinate"
    )
    earth_angle: AstropyAngles | None = Field(
        default=None, description="Angular distance between the Earth and the coordinate"
    )
    alt_az: AstropyAltAz | None = Field(default=None, description="AltAz coordinates of the coordinate")
    air_mass: NumpyArray | None = Field(default=None, description="Airmass value for the coordinate")
    sun_altitude: AstropyAngles | None = Field(
        default=None, description="Altitude of the Sun at the given time"
    )
    body_separation: dict[str, AstropyAngles] | None = Field(
        default=None, description="Angular separation from specified Solar System bodies"
    )
    body_coordinates: dict[str, AstropySkyCoords] | None = Field(
        default=None, description="Sky coordinates of specified Solar System bodies"
    )
    body_magnitude: dict[str, NumpyArray] | None = Field(
        default=None, description="Apparent magnitude of specified Solar System bodies"
    )
    galactic_bulge_separation: AstropyAngles | None = Field(
        default=None, description="Angular separation from the Galactic Bulge"
    )

    def merge(self, other: "VisibilityComputedValues") -> None:
        """
        Merge another VisibilityComputedValues into this one.

        Always updates values from the other object if they are not None.

        Parameters
        ----------
        other
            Another VisibilityComputedValues instance to merge from.

        """
        for field_name, value in other.model_dump().items():
            if value is not None:
                setattr(self, field_name, value)



from abc import ABC, abstractmethod
from datetime import datetime, timedelta

import astropy.units as u  # type: ignore[import-untyped]
import numpy as np
from astropy.constants import R_earth, R_sun  # type: ignore[import-untyped]
from astropy.coordinates import (  # type: ignore[import-untyped]
    Angle,
    EarthLocation,
    Latitude,
    Longitude,
    SkyCoord,
    get_body,
)
from astropy.time import Time, TimeDelta  # type: ignore[import-untyped]

# Define the radii of the Moon (as astropy doesn't)
R_moon = 1737.4 * u.km


class Ephemeris(ABC):
    """
    A base class for calculating ephemerides of Astronomical Observatories both
    ground and space-based. This abstract class provides the core functionality
    for computing positions and angular sizes of celestial bodies (Sun, Moon,
    Earth) relative to a spacecraft or observation point. It handles time
    series calculations with specified time intervals and step sizes.

    Parameters
    ----------
    begin : datetime or Time
        Start time of ephemeris calculation
    end : datetime or Time
        End time of ephemeris calculation
    step_size : int, TimeDelta, or timedelta, optional
        Time step between calculations in seconds, by default 60

    Attributes
    ----------
    begin : Time
        Start time of ephemeris calculation
    end : Time
        End time of ephemeris calculation
    step_size : TimeDelta
        Step size of ephemeris calculation
    timestamp : Time
        Array of calculation timestamps
    gcrs : SkyCoord
        Spacecraft position in Geocentric Celestial Reference System (GCRS)
        coordinates
    earth_location : EarthLocation
        Spacecraft position relative to Earth
    moon : SkyCoord
        Moon position relative to spacecraft
    sun : SkyCoord
        Sun position relative to spacecraft
    earth : SkyCoord
        Earth position relative to spacecraft
    longitude : Longitude, optional
        Spacecraft longitude
    latitude : Latitude, optional
        Spacecraft latitude
    height : Quantity, optional
        Spacecraft height above Earth's surface
    earth_radius_angle : Angle
        Angular radius of Earth as seen from spacecraft
    moon_radius_angle : Angle
        Angular radius of Moon as seen from spacecraft
    sun_radius_angle : Angle
        Angular radius of Sun as seen from spacecraft
    distance : Quantity
        Distance from spacecraft to Earth center

    Notes
    -----
    This is an abstract base class that must be subclassed. Subclasses must implement
    the prepare_data() method to set up the spacecraft position before ephemeris
    calculations can be performed.
    """

    # Parameters
    begin: Time  # Start time of ephemeris calculation
    end: Time  # End time of ephemeris calculation
    step_size: TimeDelta  # Step size of ephemeris calculation in seconds

    # Computed values
    timestamp: Time
    gcrs: SkyCoord
    earth_location: EarthLocation
    moon: SkyCoord
    sun: SkyCoord
    earth: SkyCoord
    longitude: Longitude | None = None
    latitude: Latitude | None = None
    height: u.Quantity | None = None
    earth_radius_angle: Angle
    moon_radius_angle: Angle
    sun_radius_angle: Angle
    distance: u.Quantity
    _start_unix: float
    _step_seconds: float

    def __init__(
        self,
        begin: datetime | Time,
        end: datetime | Time,
        step_size: int | TimeDelta | timedelta = 60,
    ) -> None:
        # Convert begin and end to astropy Time
        self.begin = begin if isinstance(begin, Time) else Time(begin)
        self.end = end if isinstance(end, Time) else Time(end)

        # Convert step_size to TimeDelta
        if isinstance(step_size, TimeDelta):
            self.step_size = step_size
        elif isinstance(step_size, timedelta):
            self.step_size = TimeDelta(step_size)
        elif isinstance(step_size, (int, float)):
            self.step_size = TimeDelta(step_size * u.s)

        # Cache scalars used frequently to avoid repeated unit conversions
        self._step_seconds = float(self.step_size.to_value(u.s))

        # Align begin/end to step grid (floor)
        s = self._step_seconds
        self.begin = Time((self.begin.unix // s) * s, format="unix")
        self.end = Time((self.end.unix // s) * s, format="unix")

        # Cache for index calculations to avoid repeated array access and conversions
        self._start_unix = float(self.begin.unix)

        # Compute range of timestamps
        self.timestamp = self._compute_timestamp()

    def __len__(self) -> int:
        return len(self.timestamp)

    def index(self, t: Time) -> int:
        """
        For a given time, return an index for the nearest time in the
        ephemeris. Note that internally converting from Time to unix makes
        this run way faster.

        Parameters
        ----------
        t : Time
            The time to find the nearest index for.

        Returns
        -------
        int
            The index of the nearest time in the ephemeris.
        """
        index = int(np.round((t.unix - self._start_unix) // self._step_seconds))
        assert index >= 0 and index < len(self), "Time outside of ephemeris of range"
        return index

    def _compute_timestamp(self) -> Time:
        """
        Get array of timestamps based on time interval and step size.

        Returns
        -------
        astropy.time.Time
            If begin equals end, returns single timestamp.
            Otherwise returns array of timestamps from begin to end with specified step_size.
        """
        return Time(
            np.arange(self.begin.unix, self.end.unix + self._step_seconds, self._step_seconds), format="unix"
        )

    def _calc(self) -> None:
        """
        Calculate ephemeris data based on the coordinates computed by
        prepare_data().
        """
        # Calculate the position of the Moon relative to the spacecraft
        self.moon = get_body("moon", self.timestamp, location=self.earth_location)

        # Calculate the position of the Sun relative to the spacecraft
        self.sun = get_body("sun", self.timestamp, location=self.earth_location)

        # Calculate the position of the Earth relative to the spacecraft
        self.earth = get_body("earth", self.timestamp, location=self.earth_location)

        # Get the longitude, latitude, height, and distance (from center of
        # Earth) of the satellite from the EarthLocation object.
        self.longitude = self.earth_location.lon
        self.latitude = self.earth_location.lat
        self.height = self.earth_location.height
        self.distance = self.gcrs.distance

        # Calculate Earth's angular radius from observatory, capped at 90 degrees
        self.earth_radius_angle = np.arcsin(np.minimum(R_earth / self.distance, 1))

        # Similarly calculate the angular radii of the Sun and the Moon, capped at 90 degrees
        self.moon_radius_angle = np.arcsin(np.minimum(R_moon / self.moon.distance, 1))
        self.sun_radius_angle = np.arcsin(np.minimum(R_sun / self.sun.distance, 1))

    @abstractmethod
    def prepare_data(self) -> None:
        """
        Prepare data for ephemeris calculation. Abstract method, to be implemented by subclasses.
        """
        raise NotImplementedError("prepare_data method must be implemented by subclass")  # pragma: no cover

    def compute(self) -> None:
        """
        Compute ephemeris data.

        This method orchestrates the computation of ephemeris data by first
        preparing the necessary data and then performing the core ephemeris
        calculations. It is intended to be called after initializing the
        Ephemeris object with the desired time range and step size.
        """
        self.prepare_data()
        self._calc()
